from copy import deepcopy
from unittest import IsolatedAsyncioTestCase
from unittest.mock import patch

from marshmallow import ValidationError

from .......messaging.decorators.attach_decorator import AttachDecorator
from .......resolver.did_resolver import DIDResolver
from .......storage.vc_holder.base import VCHolder
from .......storage.vc_holder.vc_record import VCRecord
from .......tests import mock
from .......utils.testing import create_test_profile
from .......vc.ld_proofs import DocumentLoader, DocumentVerificationResult
from .......vc.ld_proofs.constants import (
    SECURITY_CONTEXT_BBS_URL,
    SECURITY_CONTEXT_ED25519_2020_URL,
)
from .......vc.ld_proofs.error import LinkedDataProofException
from .......vc.vc_ld.manager import VcLdpManager
from .......vc.vc_ld.models.credential import VerifiableCredential
from .......vc.vc_ld.models.options import LDProofVCOptions
from .......wallet.base import BaseWallet
from .......wallet.default_verification_key_strategy import (
    BaseVerificationKeyStrategy,
    DefaultVerificationKeyStrategy,
)
from ....message_types import (
    ATTACHMENT_FORMAT,
    CRED_20_ISSUE,
    CRED_20_OFFER,
    CRED_20_PROPOSAL,
    CRED_20_REQUEST,
)
from ....messages.cred_format import V20CredFormat
from ....messages.cred_issue import V20CredIssue
from ....messages.cred_offer import V20CredOffer
from ....messages.cred_proposal import V20CredProposal
from ....messages.cred_request import V20CredRequest
from ....models.cred_ex_record import V20CredExRecord
from ....models.detail.ld_proof import V20CredExRecordLDProof
from ...handler import V20CredFormatError
from .. import handler as test_module
from ..handler import LOGGER as LD_PROOF_LOGGER
from ..handler import LDProofCredFormatHandler
from ..models.cred_detail import LDProofVCDetail
from .fixtures import (
    LD_PROOF_VC,
    LD_PROOF_VC_DETAIL,
    LD_PROOF_VC_DETAIL_BBS,
    LD_PROOF_VC_DETAIL_ED25519_2020,
)


class TestV20LDProofCredFormatHandler(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.holder = mock.MagicMock()
        self.wallet = mock.MagicMock(BaseWallet, autospec=True)

        self.profile = await create_test_profile()
        self.profile.context.injector.bind_instance(
            BaseWallet, mock.MagicMock(BaseWallet, autospec=True)
        )
        self.profile.context.injector.bind_instance(
            VCHolder, mock.MagicMock(VCHolder, autospec=True)
        )
        # Set custom document loader
        # mock_document_loader = mock.MagicMock(DocumentLoader, autospec=True)
        # mock_document_loader.custom_document_loader = custom_document_loader
        self.profile.context.injector.bind_instance(DIDResolver, DIDResolver([]))
        self.profile.context.injector.bind_instance(
            DocumentLoader, DocumentLoader(self.profile)
        )

        # Set default verkey ID strategy
        self.profile.context.injector.bind_instance(
            BaseVerificationKeyStrategy, DefaultVerificationKeyStrategy()
        )

        self.manager = VcLdpManager(self.profile)
        self.profile.context.injector.bind_instance(VcLdpManager, self.manager)
        self.handler = LDProofCredFormatHandler(self.profile)

        self.cred_proposal = V20CredProposal(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            filters_attach=[AttachDecorator.data_base64(LD_PROOF_VC_DETAIL, ident="0")],
        )

        assert self.handler.profile

    async def test_validate_fields(self):
        # Test correct data
        self.handler.validate_fields(CRED_20_PROPOSAL, LD_PROOF_VC_DETAIL)
        self.handler.validate_fields(CRED_20_OFFER, LD_PROOF_VC_DETAIL)
        self.handler.validate_fields(CRED_20_REQUEST, LD_PROOF_VC_DETAIL)
        self.handler.validate_fields(CRED_20_ISSUE, LD_PROOF_VC)

        incorrect_detail = {
            **LD_PROOF_VC_DETAIL,
            "credential": {**LD_PROOF_VC_DETAIL["credential"], "credentialSubject": None},
        }

        # test incorrect proposal
        with self.assertRaises(ValidationError):
            self.handler.validate_fields(CRED_20_PROPOSAL, incorrect_detail)

        # test incorrect offer
        with self.assertRaises(ValidationError):
            self.handler.validate_fields(CRED_20_OFFER, incorrect_detail)

        # test incorrect request
        with self.assertRaises(ValidationError):
            self.handler.validate_fields(CRED_20_REQUEST, incorrect_detail)

        # test incorrect cred
        with self.assertRaises(ValidationError):
            incorrect_cred = LD_PROOF_VC.copy()
            incorrect_cred.pop("credentialSubject")
            self.handler.validate_fields(CRED_20_ISSUE, incorrect_cred)

    async def test_get_ld_proof_detail_record(self):
        cred_ex_id = "dummy"
        details_ld_proof = [
            V20CredExRecordLDProof(
                cred_ex_id=cred_ex_id,
            ),
            V20CredExRecordLDProof(
                cred_ex_id=cred_ex_id,
            ),
        ]
        async with self.profile.session() as session:
            await details_ld_proof[0].save(session)
            await details_ld_proof[1].save(session)  # exercise logger warning on get()

        with mock.patch.object(
            LD_PROOF_LOGGER, "warning", mock.MagicMock()
        ) as mock_warning:
            assert await self.handler.get_detail_record(cred_ex_id) in details_ld_proof
            mock_warning.assert_called_once()

    async def test_create_proposal(self):
        cred_ex_record = mock.MagicMock()

        (cred_format, attachment) = await self.handler.create_proposal(
            cred_ex_record, deepcopy(LD_PROOF_VC_DETAIL)
        )

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == LD_PROOF_VC_DETAIL

        # assert data is encoded as base64
        assert attachment.data.base64

    async def test_create_proposal_adds_bbs_context(self):
        cred_ex_record = mock.MagicMock()

        (cred_format, attachment) = await self.handler.create_proposal(
            cred_ex_record, deepcopy(LD_PROOF_VC_DETAIL_BBS)
        )

        # assert BBS url added to context
        assert SECURITY_CONTEXT_BBS_URL in attachment.content["credential"]["@context"]

    async def test_create_proposal_adds_ed25519_2020_context(self):
        cred_ex_record = mock.MagicMock()

        (cred_format, attachment) = await self.handler.create_proposal(
            cred_ex_record, deepcopy(LD_PROOF_VC_DETAIL_ED25519_2020)
        )

        # assert ED25519-2020 url added to context
        assert (
            SECURITY_CONTEXT_ED25519_2020_URL
            in attachment.content["credential"]["@context"]
        )

    async def test_receive_proposal(self):
        cred_ex_record = mock.MagicMock()
        cred_proposal_message = mock.MagicMock()

        # Not much to assert. Receive proposal doesn't do anything
        await self.handler.receive_proposal(cred_ex_record, cred_proposal_message)

    async def test_create_offer(self):
        with (
            patch.object(test_module, "get_properties_without_context", return_value=[]),
        ):
            (cred_format, attachment) = await self.handler.create_offer(
                self.cred_proposal
            )

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == LD_PROOF_VC_DETAIL

        # assert data is encoded as base64
        assert attachment.data.base64

    async def test_create_offer_adds_bbs_context(self):
        cred_proposal = V20CredProposal(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            filters_attach=[
                AttachDecorator.data_base64(LD_PROOF_VC_DETAIL_BBS, ident="0")
            ],
        )

        with (
            patch.object(test_module, "get_properties_without_context", return_value=[]),
        ):
            (cred_format, attachment) = await self.handler.create_offer(cred_proposal)

        # assert BBS url added to context
        assert SECURITY_CONTEXT_BBS_URL in attachment.content["credential"]["@context"]

    async def test_create_offer_adds_ed25519_2020_context(self):
        cred_proposal = V20CredProposal(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_PROPOSAL][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            filters_attach=[
                AttachDecorator.data_base64(LD_PROOF_VC_DETAIL_ED25519_2020, ident="0")
            ],
        )

        with (
            patch.object(test_module, "get_properties_without_context", return_value=[]),
        ):
            (cred_format, attachment) = await self.handler.create_offer(cred_proposal)

        # assert BBS url added to context
        assert (
            SECURITY_CONTEXT_ED25519_2020_URL
            in attachment.content["credential"]["@context"]
        )

    async def test_create_offer_x_no_proposal(self):
        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.create_offer(None)
        assert "Cannot create linked data proof offer without proposal data" in str(
            context.exception
        )

    async def test_create_offer_x_wrong_attributes(self):
        missing_properties = ["foo"]
        with (
            patch.object(
                test_module,
                "get_properties_without_context",
                return_value=missing_properties,
            ),
            self.assertRaises(LinkedDataProofException) as context,
        ):
            await self.handler.create_offer(self.cred_proposal)

        assert (
            f"{len(missing_properties)} attributes dropped. "
            f"Provide definitions in context to correct. {missing_properties}"
            in str(context.exception)
        )

    async def test_receive_offer(self):
        cred_ex_record = mock.MagicMock()
        cred_offer_message = mock.MagicMock()

        # Not much to assert. Receive offer doesn't do anything
        await self.handler.receive_offer(cred_ex_record, cred_offer_message)

    async def test_create_bound_request(self):
        cred_offer = V20CredOffer(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(LD_PROOF_VC_DETAIL, ident="0")],
        )
        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-id",
            state=V20CredExRecord.STATE_OFFER_RECEIVED,
            cred_offer=cred_offer,
        )

        (cred_format, attachment) = await self.handler.create_request(cred_ex_record)

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == LD_PROOF_VC_DETAIL

        # assert data is encoded as base64
        assert attachment.data.base64

    async def test_create_free_request(self):
        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-id",
            state=V20CredExRecord.STATE_OFFER_RECEIVED,
            cred_proposal=self.cred_proposal,
        )

        (cred_format, attachment) = await self.handler.create_request(cred_ex_record)

        # assert identifier match
        assert cred_format.attach_id == self.handler.format.api == attachment.ident

        # assert content of attachment is proposal data
        assert attachment.content == LD_PROOF_VC_DETAIL

        # assert data is encoded as base64
        assert attachment.data.base64

    async def test_create_request_x_no_data(self):
        cred_ex_record = V20CredExRecord(state=V20CredExRecord.STATE_OFFER_RECEIVED)

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.create_request(cred_ex_record)
        assert (
            "Cannot create linked data proof request without offer or input data"
            in str(context.exception)
        )

    async def test_receive_request_no_offer(self):
        cred_ex_record = mock.MagicMock()
        cred_ex_record.cred_offer = None
        cred_request_message = mock.MagicMock()

        # Not much to assert. Receive request doesn't do anything if no prior offer
        await self.handler.receive_request(cred_ex_record, cred_request_message)

    async def test_receive_request_with_offer_no_id(self):
        cred_offer = V20CredOffer(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(LD_PROOF_VC_DETAIL, ident="0")],
        )
        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-id",
            state=V20CredExRecord.STATE_OFFER_RECEIVED,
            cred_offer=cred_offer,
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(LD_PROOF_VC_DETAIL, ident="0")],
        )

        await self.handler.receive_request(cred_ex_record, cred_request)

    async def test_receive_request_with_offer_with_id(self):
        detail = deepcopy(LD_PROOF_VC_DETAIL)
        detail["credential"]["credentialSubject"]["id"] = "some id"
        cred_offer = V20CredOffer(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(detail, ident="0")],
        )
        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-id",
            state=V20CredExRecord.STATE_OFFER_RECEIVED,
            cred_offer=cred_offer,
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(detail, ident="0")],
        )

        await self.handler.receive_request(cred_ex_record, cred_request)

    async def test_receive_request_with_offer_with_id_x_mismatch_id(self):
        detail = deepcopy(LD_PROOF_VC_DETAIL)
        detail["credential"]["credentialSubject"]["id"] = "some id"
        cred_offer = V20CredOffer(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(detail, ident="0")],
        )
        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-id",
            state=V20CredExRecord.STATE_OFFER_RECEIVED,
            cred_offer=cred_offer,
        )
        req_detail = deepcopy(detail)
        req_detail["credential"]["credentialSubject"]["id"] = "other id"
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(req_detail, ident="0")],
        )

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.receive_request(cred_ex_record, cred_request)
        assert "must match offer" in str(context.exception)

    async def test_receive_request_with_offer_with_id_x_changed_cred(self):
        detail = deepcopy(LD_PROOF_VC_DETAIL)
        cred_offer = V20CredOffer(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_OFFER][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            offers_attach=[AttachDecorator.data_base64(detail, ident="0")],
        )
        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-id",
            state=V20CredExRecord.STATE_OFFER_RECEIVED,
            cred_offer=cred_offer,
        )
        req_detail = deepcopy(LD_PROOF_VC_DETAIL_ED25519_2020)
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(req_detail, ident="0")],
        )

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.receive_request(cred_ex_record, cred_request)
        assert "Request must match offer if offer is sent" in str(context.exception)

    async def test_issue_credential(self):
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(LD_PROOF_VC_DETAIL, ident="0")],
        )

        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_request=cred_request,
        )

        with mock.patch.object(
            VcLdpManager,
            "issue",
            mock.CoroutineMock(
                return_value=VerifiableCredential.deserialize(LD_PROOF_VC)
            ),
        ) as mock_issue:
            (cred_format, attachment) = await self.handler.issue_credential(
                cred_ex_record
            )

            LDProofVCDetail.deserialize(LD_PROOF_VC_DETAIL)

            mock_issue.assert_called_once_with(
                VerifiableCredential.deserialize(LD_PROOF_VC_DETAIL["credential"]),
                LDProofVCOptions.deserialize(LD_PROOF_VC_DETAIL["options"]),
            )

            # assert identifier match
            assert cred_format.attach_id == self.handler.format.api == attachment.ident

            # assert content of attachment is credential data
            assert attachment.content == LD_PROOF_VC

            # assert data is encoded as base64
            assert attachment.data.base64

    async def test_issue_credential_x_no_data(self):
        cred_ex_record = V20CredExRecord()

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.issue_credential(cred_ex_record)
        assert "Cannot issue credential without credential request" in str(
            context.exception
        )

    async def test_receive_credential(self):
        cred_issue = V20CredIssue(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_ISSUE][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            credentials_attach=[AttachDecorator.data_base64(LD_PROOF_VC, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(LD_PROOF_VC_DETAIL, ident="0")],
        )
        cred_ex_record = V20CredExRecord(
            cred_ex_id="cred-ex-id",
            cred_request=cred_request,
        )

        await self.handler.receive_credential(cred_ex_record, cred_issue)

    async def test_receive_credential_x_credential_ne_request(self):
        detail = deepcopy(LD_PROOF_VC_DETAIL)

        # Change date so request is different than issued credential
        detail["credential"]["issuanceDate"] = "2020-01-01"

        cred_issue = V20CredIssue(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_ISSUE][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            credentials_attach=[AttachDecorator.data_base64(LD_PROOF_VC, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(detail, ident="0")],
        )
        cred_ex_record = V20CredExRecord(
            cred_ex_id="cred-ex-id",
            cred_request=cred_request,
        )

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.receive_credential(cred_ex_record, cred_issue)
        assert "does not match requested credential" in str(context.exception)

    async def test_receive_credential_x_credential_status_ne(self):
        detail = deepcopy(LD_PROOF_VC_DETAIL)

        # Set credential status so it's only set on the detail
        # not the issued credential
        detail["options"]["credentialStatus"] = {"type": "CredentialStatusType"}

        cred_issue = V20CredIssue(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_ISSUE][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            credentials_attach=[AttachDecorator.data_base64(LD_PROOF_VC, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(detail, ident="0")],
        )
        cred_ex_record = V20CredExRecord(
            cred_ex_id="cred-ex-id",
            cred_request=cred_request,
        )

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.receive_credential(cred_ex_record, cred_issue)
        assert "Received credential status contains credential status" in str(
            context.exception
        )

    async def test_receive_credential_x_credential_status_ne_both_set(self):
        detail = deepcopy(LD_PROOF_VC_DETAIL)
        status_entry = {"type": "SomeRandomType"}

        # Set credential status in both request and reference credential
        detail["options"]["credentialStatus"] = {"type": "CredentialStatusType"}
        detail["credential"]["credentialStatus"] = deepcopy(status_entry)

        vc = deepcopy(LD_PROOF_VC)
        vc["credentialStatus"] = deepcopy(status_entry)

        cred_issue = V20CredIssue(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_ISSUE][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            credentials_attach=[AttachDecorator.data_base64(vc, ident="0")],
        )
        cred_request = V20CredRequest(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            requests_attach=[AttachDecorator.data_base64(detail, ident="0")],
        )
        cred_ex_record = V20CredExRecord(
            cred_ex_id="cred-ex-id",
            cred_request=cred_request,
        )

        with self.assertRaises(V20CredFormatError) as context:
            await self.handler.receive_credential(cred_ex_record, cred_issue)
        assert "Received credential status type does not match credential request" in str(
            context.exception
        )

    async def test_receive_credential_x_proof_options_ne(self):
        properties = {
            "challenge": "3f9054c0-70df-497d-9bbb-f373ddf986ce",
            "domain": "example.com",
            "proofType": "SomeType",
            "created": "2000-01-11T03:50:55",
        }
        for property, value in properties.items():
            detail = deepcopy(LD_PROOF_VC_DETAIL)

            detail["options"][property] = value

            cred_issue = V20CredIssue(
                formats=[
                    V20CredFormat(
                        attach_id="0",
                        format_=ATTACHMENT_FORMAT[CRED_20_ISSUE][
                            V20CredFormat.Format.LD_PROOF.api
                        ],
                    )
                ],
                credentials_attach=[AttachDecorator.data_base64(LD_PROOF_VC, ident="0")],
            )
            cred_request = V20CredRequest(
                formats=[
                    V20CredFormat(
                        attach_id="0",
                        format_=ATTACHMENT_FORMAT[CRED_20_REQUEST][
                            V20CredFormat.Format.LD_PROOF.api
                        ],
                    )
                ],
                requests_attach=[AttachDecorator.data_base64(detail, ident="0")],
            )
            cred_ex_record = V20CredExRecord(
                cred_ex_id="cred-ex-id",
                cred_request=cred_request,
            )

            with self.assertRaises(V20CredFormatError) as context:
                await self.handler.receive_credential(cred_ex_record, cred_issue)
            assert f"does not match options.{property} from credential request" in str(
                context.exception
            )

    async def test_store_credential(self):
        cred_issue = V20CredIssue(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_ISSUE][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            credentials_attach=[AttachDecorator.data_base64(LD_PROOF_VC, ident="0")],
        )

        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_issue=cred_issue,
        )

        cred_id = "cred_id"
        self.holder.store_credential = mock.CoroutineMock()

        with mock.patch.object(
            VcLdpManager,
            "verify_credential",
            mock.CoroutineMock(return_value=DocumentVerificationResult(verified=True)),
        ):
            await self.handler.store_credential(cred_ex_record, cred_id)

            self.profile.context.injector.get_provider(
                VCHolder
            )._instance.store_credential.assert_called_once_with(
                VCRecord(
                    contexts=LD_PROOF_VC["@context"],
                    expanded_types=[
                        "https://www.w3.org/2018/credentials#VerifiableCredential",
                        "https://example.org/examples#UniversityDegreeCredential",
                    ],
                    issuer_id=LD_PROOF_VC["issuer"],
                    subject_ids=[],
                    schema_ids=[],  # Schemas not supported yet
                    proof_types=[LD_PROOF_VC["proof"]["type"]],
                    cred_value=LD_PROOF_VC,
                    given_id=None,
                    record_id=cred_id,
                )
            )

    async def test_store_credential_x_not_verified(self):
        cred_issue = V20CredIssue(
            formats=[
                V20CredFormat(
                    attach_id="0",
                    format_=ATTACHMENT_FORMAT[CRED_20_ISSUE][
                        V20CredFormat.Format.LD_PROOF.api
                    ],
                )
            ],
            credentials_attach=[AttachDecorator.data_base64(LD_PROOF_VC, ident="0")],
        )

        cred_ex_record = V20CredExRecord(
            cred_ex_id="dummy-cxid",
            cred_issue=cred_issue,
        )

        cred_id = "cred_id"
        self.holder.store_credential = mock.CoroutineMock()

        with (
            mock.patch.object(
                self.manager,
                "_get_suite",
                mock.CoroutineMock(),
            ),
            mock.patch.object(
                self.manager,
                "verify_credential",
                mock.CoroutineMock(
                    return_value=DocumentVerificationResult(verified=False)
                ),
            ),
            mock.patch.object(
                self.manager,
                "_get_proof_purpose",
            ),
            self.assertRaises(V20CredFormatError) as context,
        ):
            await self.handler.store_credential(cred_ex_record, cred_id)
        assert "Received invalid credential: " in str(context.exception)



"""Validators for schema fields."""

import json
import re

from base58 import alphabet
from marshmallow.exceptions import ValidationError
from marshmallow.fields import Field
from marshmallow.validate import OneOf, Range, Regexp, Validator

from ..ledger.endpoint_type import EndpointType as EndpointTypeEnum
from ..revocation.models.revocation_registry import RevocationRegistry
from ..wallet.did_posture import DIDPosture as DIDPostureEnum
from .util import epoch_to_str

B58 = alphabet if isinstance(alphabet, str) else alphabet.decode("ascii")

EXAMPLE_TIMESTAMP = 1640995199  # 2021-12-31 23:59:59Z


class StrOrDictField(Field):
    """URI or Dict field for Marshmallow."""

    def _deserialize(self, value, attr, data, **kwargs):
        if not isinstance(value, (str, dict)):
            raise ValidationError("Field should be str or dict")
        return super()._deserialize(value, attr, data, **kwargs)


class StrOrNumberField(Field):
    """String or Number field for Marshmallow."""

    def _deserialize(self, value, attr, data, **kwargs):
        if not isinstance(value, (str, float, int)):
            raise ValidationError("Field should be str or int or float")
        return super()._deserialize(value, attr, data, **kwargs)


class DictOrDictListField(Field):
    """Dict or Dict List field for Marshmallow."""

    def _deserialize(self, value, attr, data, **kwargs):
        if not isinstance(value, dict):
            if not isinstance(value, list) or not all(
                isinstance(item, dict) for item in value
            ):
                raise ValidationError("Field should be dict or list of dicts")
        return super()._deserialize(value, attr, data, **kwargs)


class UriOrDictField(StrOrDictField):
    """URI or Dict field for Marshmallow."""

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, str):
            # Check regex
            Uri()(value)
        return super()._deserialize(value, attr, data, **kwargs)


class IntEpoch(Range):
    """Validate value against (integer) epoch format."""

    EXAMPLE = EXAMPLE_TIMESTAMP

    def __init__(self):
        """Initialize the instance."""

        super().__init__(  # use u64 for indy-sdk compatibility
            min=0,
            max=18446744073709551615,
            error="Value {input} is not a valid integer epoch time",
        )


class WholeNumber(Range):
    """Validate value as non-negative integer."""

    EXAMPLE = 0

    def __init__(self):
        """Initialize the instance."""

        super().__init__(min=0, error="Value {input} is not a non-negative integer")

    def __call__(self, value):
        """Validate input value."""

        if not isinstance(value, int):
            raise ValidationError("Value {input} is not a valid whole number")
        super().__call__(value)


class NumericStrWhole(Regexp):
    """Validate value against whole number numeric string."""

    EXAMPLE = "0"
    PATTERN = r"^[0-9]*$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            NumericStrWhole.PATTERN,
            error="Value {input} is not a non-negative numeric string",
        )


class NumericStrAny(Regexp):
    """Validate value against any number numeric string."""

    EXAMPLE = "-1"
    PATTERN = r"^-?[0-9]*$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            NumericStrAny.PATTERN,
            error="Value {input} is not a numeric string",
        )


class NaturalNumber(Range):
    """Validate value as positive integer."""

    EXAMPLE = 10

    def __init__(self):
        """Initialize the instance."""

        super().__init__(min=1, error="Value {input} is not a positive integer")

    def __call__(self, value):
        """Validate input value."""

        if not isinstance(value, int):
            raise ValidationError("Value {input} is not a valid natural number")
        super().__call__(value)


class NumericStrNatural(Regexp):
    """Validate value against natural number numeric string."""

    EXAMPLE = "1"
    PATTERN = r"^[1-9][0-9]*$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            NumericStrNatural.PATTERN,
            error="Value {input} is not a positive numeric string",
        )


class IndyRevRegSize(Range):
    """Validate value as indy revocation registry size."""

    EXAMPLE = 1000

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            min=RevocationRegistry.MIN_SIZE,
            max=RevocationRegistry.MAX_SIZE,
            error=(
                "Value {input} must be an integer between "
                f"{RevocationRegistry.MIN_SIZE} and "
                f"{RevocationRegistry.MAX_SIZE} inclusively"
            ),
        )

    def __call__(self, value):
        """Validate input value."""

        if not isinstance(value, int):
            raise ValidationError(
                "Value {input} must be an integer between "
                f"{RevocationRegistry.MIN_SIZE} and "
                f"{RevocationRegistry.MAX_SIZE} inclusively"
            )
        super().__call__(value)


class JWSHeaderKid(Regexp):
    """Validate value against JWS header kid."""

    EXAMPLE = "did:sov:LjgpST2rjsoxYegQDRm7EL#keys-4"
    PATTERN = rf"^did:(?:key:z[{B58}]+|sov:[{B58}]{{21,22}}(;.*)?(\?.*)?#.+)$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            JWSHeaderKid.PATTERN,
            error="Value {input} is neither in W3C did:key nor DID URL format",
        )


class NonSDList(Regexp):
    """Validate NonSD List."""

    EXAMPLE = [
        "name",
        "address",
        "address.street_address",
        "nationalities[1:3]",
    ]
    PATTERN = r"[a-z0-9:\[\]_\.@?\(\)]"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            NonSDList.PATTERN,
            error="Value {input} is not a valid NonSDList",
        )


class JSONWebToken(Regexp):
    """Validate JSON Web Token."""

    EXAMPLE = (
        "eyJhbGciOiJFZERTQSJ9.eyJhIjogIjAifQ.dBjftJeZ4CVP-mB92K27uhbUJU1p1r_wW1gFWFOEjXk"
    )
    PATTERN = r"^[a-zA-Z0-9_-]+\.[a-zA-Z0-9_-]*\.[a-zA-Z0-9_-]+$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            JSONWebToken.PATTERN,
            error="Value {input} is not a valid JSON Web token",
        )


class SDJSONWebToken(Regexp):
    """Validate SD-JSON Web Token."""

    EXAMPLE = (
        "eyJhbGciOiJFZERTQSJ9."
        "eyJhIjogIjAifQ."
        "dBjftJeZ4CVP-mB92K27uhbUJU1p1r_wW1gFWFOEjXk"
        "~WyJEM3BUSFdCYWNRcFdpREc2TWZKLUZnIiwgIkRFIl0"
        "~WyJPMTFySVRjRTdHcXExYW9oRkd0aDh3IiwgIlNBIl0"
        "~WyJkVmEzX1JlTGNsWTU0R1FHZm5oWlRnIiwgInVwZGF0ZWRfYXQiLCAxNTcwMDAwMDAwXQ"
    )
    PATTERN = r"^[a-zA-Z0-9_-]+\.[a-zA-Z0-9_-]*\.[a-zA-Z0-9_-]+(?:~[a-zA-Z0-9._-]+)*~?$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            SDJSONWebToken.PATTERN,
            error="Value {input} is not a valid SD-JSON Web token",
        )


class DIDKey(Regexp):
    """Validate value against DID key specification."""

    EXAMPLE = "did:key:z6MkpTHR8VNsBxYAAWHut2Geadd9jSwuBV8xRoAnwWsdvktH"
    PATTERN = re.compile(rf"^did:key:z[{B58}]+$")

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            DIDKey.PATTERN, error="Value {input} is not in W3C did:key format"
        )


class DIDKeyOrRef(Regexp):
    """Validate value against DID key specification."""

    EXAMPLE = "did:key:z6MkpTHR8VNsBxYAAWHut2Geadd9jSwuBV8xRoAnwWsdvktH"
    PATTERN = re.compile(rf"^did:key:z[{B58}]+(?:#z[{B58}]+)?$")

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            DIDKeyOrRef.PATTERN, error="Value {input} is not a did:key or did:key ref"
        )


class DIDKeyRef(Regexp):
    """Validate value as DID key reference."""

    EXAMPLE = (
        "did:key:z6MkpTHR8VNsBxYAAWHut2Geadd9jSwuBV8xRoAnwWsdvktH"
        "#z6MkpTHR8VNsBxYAAWHut2Geadd9jSwuBV8xRoAnwWsdvktH"
    )
    PATTERN = re.compile(rf"^did:key:z[{B58}]+#z[{B58}]+$")

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            DIDKeyRef.PATTERN, error="Value {input} is not a did:key reference"
        )


class DIDWeb(Regexp):
    """Validate value against did:web specification."""

    EXAMPLE = "did:web:example.com"
    PATTERN = re.compile(r"^(did:web:)([a-zA-Z0-9%._-]*:)*[a-zA-Z0-9%._-]+$")

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            DIDWeb.PATTERN, error="Value {input} is not in W3C did:web format"
        )


class DIDWebvh(Regexp):
    """Validate value against did:webvh specification."""

    EXAMPLE = (
        "did:webvh:QmP9VWaTCHcyztDpRj9XSHvZbmYe3m9HZ61KoDtZgWaXVU:example.com%3A5000"
    )
    PATTERN = re.compile(r"^(did:webvh:)([a-zA-Z0-9%._-]*:)*[a-zA-Z0-9%._-]+$")

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            DIDWebvh.PATTERN, error="Value {input} is not in W3C did:webvh format"
        )


class DIDPosture(OneOf):
    """Validate value against defined DID postures."""

    EXAMPLE = DIDPostureEnum.WALLET_ONLY.moniker

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            choices=[did_posture.moniker for did_posture in DIDPostureEnum],
            error="Value {input} must be one of {choices}",
        )


class IndyDID(Regexp):
    """Validate value against indy DID."""

    EXAMPLE = "did:indy:sovrin:WRfXPg8dantKVubE3HX8pw"
    PATTERN = re.compile(rf"^(did:(sov|indy):)?[{B58}]{{21,22}}$")

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            IndyDID.PATTERN,
            error="Value {input} is not an indy decentralized identifier (DID)",
        )


class AnonCredsDID(Regexp):
    """Validate value against anoncreds DID."""

    METHOD = r"([a-zA-Z0-9_]+)"
    NETWORK = r"(:[a-zA-Z0-9_.%-]+)?"  # Optional network
    METHOD_ID = r"([a-zA-Z0-9_.%-]+(:[a-zA-Z0-9_.%-]+)*)"

    EXAMPLE = "did:(method):WgWxqztrNooG92RXvxSTWv"
    PATTERN = re.compile(rf"^did:{METHOD}{NETWORK}:{METHOD_ID}")

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            DIDValidation.PATTERN,
            error="Value {input} is not an decentralized identifier (DID)",
        )


class DIDValidation(Regexp):
    """Validate value against any valid DID spec."""

    METHOD = r"([a-zA-Z0-9_]+)"
    NETWORK = r"(:[a-zA-Z0-9_.%-]+)?"  # Optional network
    METHOD_ID = r"([a-zA-Z0-9_.%-]+(:[a-zA-Z0-9_.%-]+)*)"
    PARAMS = r"((;[a-zA-Z0-9_.:%-]+=[a-zA-Z0-9_.:%-]*)*)"
    PATH = r"(\/[^#?]*)?"
    QUERY = r"([?][^#]*)?"
    FRAGMENT = r"(\#.*)?$"

    EXAMPLE = "did:peer:WgWxqztrNooG92RXvxSTWv"
    PATTERN = re.compile(
        rf"^did:{METHOD}{NETWORK}:{METHOD_ID}{PARAMS}{PATH}{QUERY}{FRAGMENT}$"
    )

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            DIDValidation.PATTERN,
            error="Value {input} is not a valid DID",
        )


# temporary support for short Indy DIDs in place of qualified DIDs
class MaybeIndyDID(Regexp):
    """Validate value against any valid DID spec or a short Indy DID."""

    EXAMPLE = DIDValidation.EXAMPLE
    PATTERN = re.compile(IndyDID.PATTERN.pattern + "|" + DIDValidation.PATTERN.pattern)

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            MaybeIndyDID.PATTERN,
            error="Value {input} is not a valid DID",
        )


class RawPublicEd25519VerificationKey2018(Regexp):
    """Validate value against (Ed25519VerificationKey2018) raw public key."""

    EXAMPLE = "H3C2AVvLMv6gmMNam3uVAjZpfkcJCwDwnZn6z3wXmqPV"
    PATTERN = rf"^[{B58}]{{43,44}}$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            RawPublicEd25519VerificationKey2018.PATTERN,
            error="Value {input} is not a raw Ed25519VerificationKey2018 key",
        )


class RoutingKey(Regexp):
    """Validate between indy or did key.

    Validate value against indy (Ed25519VerificationKey2018)
    raw public key or DID key specification.
    """

    EXAMPLE = DIDKey.EXAMPLE
    PATTERN = re.compile(
        DIDKey.PATTERN.pattern + "|" + RawPublicEd25519VerificationKey2018.PATTERN
    )

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            RoutingKey.PATTERN,
            error=(
                "Value {input} is not in W3C did:key"
                " or Ed25519VerificationKey2018 key format"
            ),
        )


class IndyCredDefId(Regexp):
    """Validate value against indy credential definition identifier specification."""

    EXAMPLE = "WgWxqztrNooG92RXvxSTWv:3:CL:20:tag"
    PATTERN = (
        rf"^([{B58}]{{21,22}})"  # issuer DID
        f":3"  # cred def id marker
        f":CL"  # sig alg
        rf":(([1-9][0-9]*)|([{B58}]{{21,22}}:2:.+:[0-9.]+))"  # schema txn / id
        f":(.+)?$"  # tag
    )

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            IndyCredDefId.PATTERN,
            error="Value {input} is not an indy credential definition identifier",
        )


class AnonCredsCredDefId(Regexp):
    """Validate value against anoncreds credential definition identifier specification."""

    EXAMPLE = "did:(method):3:CL:20:tag"
    PATTERN = r"^(.+$)"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            AnonCredsCredDefId.PATTERN,
            error="Value {input} is not an anoncreds credential definition identifier",
        )


class MajorMinorVersion(Regexp):
    """Validate value against major minor version specification."""

    EXAMPLE = "1.0"
    PATTERN = r"^[0-9.]+$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            MajorMinorVersion.PATTERN,
            error="Value {input} is not a valid version major minor version (use only digits and '.')",  # noqa: E501
        )


class IndySchemaId(Regexp):
    """Validate value against indy schema identifier specification."""

    EXAMPLE = "WgWxqztrNooG92RXvxSTWv:2:schema_name:1.0"
    PATTERN = rf"^[{B58}]{{21,22}}:2:.+:[0-9.]+$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            IndySchemaId.PATTERN,
            error="Value {input} is not an indy schema identifier",
        )


class AnonCredsSchemaId(Regexp):
    """Validate value against indy schema identifier specification."""

    EXAMPLE = "did:(method):2:schema_name:1.0"
    PATTERN = r"^(.+$)"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            AnonCredsSchemaId.PATTERN,
            error="Value {input} is not an anoncreds schema identifier",
        )


class IndyRevRegId(Regexp):
    """Validate value against indy revocation registry identifier specification."""

    EXAMPLE = "WgWxqztrNooG92RXvxSTWv:4:WgWxqztrNooG92RXvxSTWv:3:CL:20:tag:CL_ACCUM:0"
    PATTERN = (
        rf"^([{B58}]{{21,22}}):4:"
        rf"([{B58}]{{21,22}}):3:"
        rf"CL:(([1-9][0-9]*)|([{B58}]{{21,22}}:2:.+:[0-9.]+))(:.+)?:"
        rf"CL_ACCUM:(.+$)"
    )

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            IndyRevRegId.PATTERN,
            error="Value {input} is not an indy revocation registry identifier",
        )


class AnonCredsRevRegId(Regexp):
    """Validate value against anoncreds revocation registry identifier specification."""

    EXAMPLE = "did:(method):4:did:<method>:3:CL:20:tag:CL_ACCUM:0"
    PATTERN = r"^(.+$)"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            AnonCredsRevRegId.PATTERN,
            error="Value {input} is not an anoncreds revocation registry identifier",
        )


class IndyCredRevId(Regexp):
    """Validate value against indy credential revocation identifier specification."""

    EXAMPLE = "12345"
    PATTERN = r"^[1-9][0-9]*$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            IndyCredRevId.PATTERN,
            error="Value {input} is not an indy credential revocation identifier",
        )


class AnonCredsCredRevId(Regexp):
    """Validate value against anoncreds credential revocation identifier specification."""

    EXAMPLE = "12345"
    PATTERN = r"^[1-9][0-9]*$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            AnonCredsCredRevId.PATTERN,
            error="Value {input} is not an anoncreds credential revocation identifier",
        )


class Predicate(OneOf):
    """Validate value against predicate."""

    EXAMPLE = ">="

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            choices=["<", "<=", ">=", ">"],
            error="Value {input} must be one of {choices}",
        )


class ISO8601DateTime(Regexp):
    """Validate value against ISO 8601 datetime format."""

    EXAMPLE = epoch_to_str(EXAMPLE_TIMESTAMP)
    PATTERN = (
        r"^\d{4}-\d\d-\d\d[T ]\d\d:\d\d"
        r"(?:\:(?:\d\d(?:\.\d{1,6})?))?(?:[+-]\d\d:?\d\d|Z|)$"
    )

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            ISO8601DateTime.PATTERN,
            error="Value {input} is not a date in valid format",
        )


class RFC3339DateTime(Regexp):
    """Validate value against RFC3339 datetime format."""

    EXAMPLE = "2010-01-01T19:23:24Z"
    PATTERN = (
        r"^([0-9]{4})-([0-9]{2})-([0-9]{2})([Tt ]([0-9]{2}):([0-9]{2}):"
        r"([0-9]{2})(\.[0-9]+)?)?(([Zz]|([+-])([0-9]{2}):([0-9]{2})))?$"
    )

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            RFC3339DateTime.PATTERN,
            error="Value {input} is not a date in valid format",
        )


class IndyWQL(Regexp):  # using Regexp brings in nice visual validator cue
    """Validate value as potential WQL query."""

    EXAMPLE = json.dumps({"attr::name::value": "Alex"})
    PATTERN = r"^{.*}$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            IndyWQL.PATTERN,
            error="Value {input} is not a valid WQL query",
        )

    def __call__(self, value):
        """Validate input value."""

        super().__call__(value or "")
        message = f"Value {value} is not a valid WQL query"

        try:
            json.loads(value)
        except (json.JSONDecodeError, TypeError):
            raise ValidationError(message)

        return value


class IndyExtraWQL(Regexp):  # using Regexp brings in nice visual validator cue
    """Validate value as potential extra WQL query in cred search for proof req."""

    EXAMPLE = json.dumps({"0_drink_uuid": {"attr::drink::value": "martini"}})
    PATTERN = r'^{\s*".*?"\s*:\s*{.*?}\s*(,\s*".*?"\s*:\s*{.*?}\s*)*\s*}$'

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            IndyExtraWQL.PATTERN,
            error="Value {input} is not a valid extra WQL query",
        )

    def __call__(self, value):
        """Validate input value."""

        super().__call__(value or "")
        message = f"Value {value} is not a valid extra WQL query"

        try:
            json.loads(value)
        except (json.JSONDecodeError, TypeError):
            raise ValidationError(message)

        return value


class Base64(Regexp):
    """Validate base64 value."""

    EXAMPLE = "ey4uLn0="
    PATTERN = r"^[a-zA-Z0-9+/]*={0,2}$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            Base64.PATTERN,
            error="Value {input} is not a valid base64 encoding",
        )


class Base64URL(Regexp):
    """Validate base64 value."""

    EXAMPLE = "ey4uLn0="
    PATTERN = r"^[-_a-zA-Z0-9]*={0,2}$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            Base64URL.PATTERN,
            error="Value {input} is not a valid base64url encoding",
        )


class Base64URLNoPad(Regexp):
    """Validate base64 value."""

    EXAMPLE = "ey4uLn0"
    PATTERN = r"^[-_a-zA-Z0-9]*$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            Base64URLNoPad.PATTERN,
            error="Value {input} is not a valid unpadded base64url encoding",
        )


class SHA256Hash(Regexp):
    """Validate (binhex-encoded) SHA256 value."""

    EXAMPLE = "617a48c7c8afe0521efdc03e5bb0ad9e655893e6b4b51f0e794d70fba132aacb"
    PATTERN = r"^[a-fA-F0-9+/]{64}$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            SHA256Hash.PATTERN,
            error="Value {input} is not a valid (binhex-encoded) SHA-256 hash",
        )


class Base58SHA256Hash(Regexp):
    """Validate value against base58 encoding of SHA-256 hash."""

    EXAMPLE = "H3C2AVvLMv6gmMNam3uVAjZpfkcJCwDwnZn6z3wXmqPV"
    PATTERN = rf"^[{B58}]{{43,44}}$"

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            Base58SHA256Hash.PATTERN,
            error="Value {input} is not a base58 encoding of a SHA-256 hash",
        )


class UUIDFour(Regexp):
    """Validate UUID4: 8-4-4-4-12 hex digits, the 13th of which being 4."""

    EXAMPLE = "3fa85f64-5717-4562-b3fc-2c963f66afa6"
    PATTERN = (
        r"[a-fA-F0-9]{8}-"
        r"[a-fA-F0-9]{4}-"
        r"4[a-fA-F0-9]{3}-"
        r"[a-fA-F0-9]{4}-"
        r"[a-fA-F0-9]{12}"
    )

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            UUIDFour.PATTERN,
            error="Value {input} is not UUID4 (8-4-4-4-12 hex digits with digit#13=4)",
        )


class Uri(Regexp):
    """Validate value against URI on any scheme."""

    EXAMPLE = "https://www.w3.org/2018/credentials/v1"
    PATTERN = r"\w+:(\/?\/?)[^\s]+"

    def __init__(self):
        """Initialize the instance."""
        super().__init__(Uri.PATTERN, error="Value {input} is not URI")


class Endpoint(Regexp):  # using Regexp brings in nice visual validator cue
    """Validate value against endpoint URL on any scheme."""

    EXAMPLE = "https://myhost:8021"
    PATTERN = (
        r"^[A-Za-z0-9\.\-\+]+:"  # scheme
        r"//([A-Za-z0-9][.A-Za-z0-9-_]+[A-Za-z0-9])+"  # host
        r"(:[1-9][0-9]*)?"  # port
        r"(/[^?&#]+)?$"  # path
    )

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            Endpoint.PATTERN,
            error="Value {input} is not a valid endpoint",
        )


class EndpointType(OneOf):
    """Validate value against allowed endpoint/service types."""

    EXAMPLE = EndpointTypeEnum.ENDPOINT.w3c

    def __init__(self):
        """Initialize the instance."""

        super().__init__(
            choices=[e.w3c for e in EndpointTypeEnum],
            error="Value {input} must be one of {choices}",
        )


class CredentialType(Validator):
    """Credential Type."""

    CREDENTIAL_TYPE = "VerifiableCredential"
    EXAMPLE = [CREDENTIAL_TYPE, "AlumniCredential"]

    def __init__(self) -> None:
        """Initialize the instance."""
        super().__init__()

    def __call__(self, value):
        """Validate input value."""
        length = len(value)
        if length < 1 or CredentialType.CREDENTIAL_TYPE not in value:
            raise ValidationError(f"type must include {CredentialType.CREDENTIAL_TYPE}")

        return value


class PresentationType(Validator):
    """Presentation Type."""

    PRESENTATION_TYPE = "VerifiablePresentation"
    EXAMPLE = [PRESENTATION_TYPE]

    def __init__(self) -> None:
        """Initialize the instance."""
        super().__init__()

    def __call__(self, value):
        """Validate input value."""
        length = len(value)
        if length < 1 or PresentationType.PRESENTATION_TYPE not in value:
            raise ValidationError(
                f"type must include {PresentationType.PRESENTATION_TYPE}"
            )

        return value


class CredentialContext(Validator):
    """Credential Context."""

    FIRST_CONTEXT = "https://www.w3.org/2018/credentials/v1"
    VALID_CONTEXTS = [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/ns/credentials/v2",
    ]
    EXAMPLE = [VALID_CONTEXTS[0], "https://www.w3.org/2018/credentials/examples/v1"]

    def __init__(self) -> None:
        """Initialize the instance."""
        super().__init__()

    def __call__(self, value):
        """Validate input value."""

        if not isinstance(value, list):
            raise ValidationError("Value must be a non-empty list.")

        if not value or value[0] not in CredentialContext.VALID_CONTEXTS:
            raise ValidationError(
                f"First context must be one of {CredentialContext.VALID_CONTEXTS}"
            )

        return value


class CredentialSubject(Validator):
    """Credential subject."""

    EXAMPLE = {
        "id": "did:example:ebfeb1f712ebc6f1c276e12ec21",
        "alumniOf": {"id": "did:example:c276e12ec21ebfeb1f712ebc6f1"},
    }

    def __init__(self) -> None:
        """Initialize the instance."""
        super().__init__()

    def __call__(self, value):
        """Validate input value."""
        subjects = value if isinstance(value, list) else [value]

        for subject in subjects:
            if "id" in subject:
                uri_validator = Uri()
                try:
                    uri_validator(subject["id"])
                except ValidationError:
                    raise ValidationError(
                        f"credential subject id {subject['id']} must be URI"
                    ) from None

        return value


class CredentialStatus(Validator):
    """Credential status."""

    EXAMPLE = {
        "id": "https://example.com/credentials/status/3#94567",
        "type": "BitstringStatusListEntry",
        "statusPurpose": "revocation",
        "statusListIndex": "94567",
        "statusListCredential": "https://example.com/credentials/status/3",
    }

    def __init__(self) -> None:
        """Initialize the instance."""
        super().__init__()

    def __call__(self, value):
        """Validate input value."""
        # TODO write some tests

        return value


class IndyOrKeyDID(Regexp):
    """Indy or Key DID class."""

    PATTERN = "|".join(x.pattern for x in [DIDKey.PATTERN, IndyDID.PATTERN])
    EXAMPLE = IndyDID.EXAMPLE

    def __init__(
        self,
    ):
        """Initialize the instance."""
        super().__init__(
            IndyOrKeyDID.PATTERN,
            error="Value {input} is not in did:key or indy did format",
        )


# Instances for marshmallow schema specification
INT_EPOCH_VALIDATE = IntEpoch()
INT_EPOCH_EXAMPLE = IntEpoch.EXAMPLE

WHOLE_NUM_VALIDATE = WholeNumber()
WHOLE_NUM_EXAMPLE = WholeNumber.EXAMPLE

NUM_STR_WHOLE_VALIDATE = NumericStrWhole()
NUM_STR_WHOLE_EXAMPLE = NumericStrWhole.EXAMPLE

NUM_STR_ANY_VALIDATE = NumericStrAny()
NUM_STR_ANY_EXAMPLE = NumericStrAny.EXAMPLE

NATURAL_NUM_VALIDATE = NaturalNumber()
NATURAL_NUM_EXAMPLE = NaturalNumber.EXAMPLE

NUM_STR_NATURAL_VALIDATE = NumericStrNatural()
NUM_STR_NATURAL_EXAMPLE = NumericStrNatural.EXAMPLE

INDY_REV_REG_SIZE_VALIDATE = IndyRevRegSize()
INDY_REV_REG_SIZE_EXAMPLE = IndyRevRegSize.EXAMPLE

JWS_HEADER_KID_VALIDATE = JWSHeaderKid()
JWS_HEADER_KID_EXAMPLE = JWSHeaderKid.EXAMPLE

NON_SD_LIST_VALIDATE = NonSDList()
NON_SD_LIST_EXAMPLE = NonSDList().EXAMPLE

JWT_VALIDATE = JSONWebToken()
JWT_EXAMPLE = JSONWebToken.EXAMPLE

SD_JWT_VALIDATE = SDJSONWebToken()
SD_JWT_EXAMPLE = SDJSONWebToken.EXAMPLE

DID_KEY_VALIDATE = DIDKey()
DID_KEY_EXAMPLE = DIDKey.EXAMPLE

DID_KEY_OR_REF_VALIDATE = DIDKeyOrRef()
DID_KEY_OR_REF_EXAMPLE = DIDKeyOrRef.EXAMPLE

DID_KEY_REF_VALIDATE = DIDKeyRef()
DID_KEY_REF_EXAMPLE = DIDKeyRef.EXAMPLE

DID_POSTURE_VALIDATE = DIDPosture()
DID_POSTURE_EXAMPLE = DIDPosture.EXAMPLE

DID_WEB_VALIDATE = DIDWeb()
DID_WEB_EXAMPLE = DIDWeb.EXAMPLE

DID_WEBVH_VALIDATE = DIDWebvh()
DID_WEBVH_EXAMPLE = DIDWebvh.EXAMPLE

ROUTING_KEY_VALIDATE = RoutingKey()
ROUTING_KEY_EXAMPLE = RoutingKey.EXAMPLE

INDY_DID_VALIDATE = IndyDID()
INDY_DID_EXAMPLE = IndyDID.EXAMPLE

GENERIC_DID_VALIDATE = MaybeIndyDID()
GENERIC_DID_EXAMPLE = MaybeIndyDID.EXAMPLE

RAW_ED25519_2018_PUBLIC_KEY_VALIDATE = RawPublicEd25519VerificationKey2018()
RAW_ED25519_2018_PUBLIC_KEY_EXAMPLE = RawPublicEd25519VerificationKey2018.EXAMPLE

INDY_SCHEMA_ID_VALIDATE = IndySchemaId()
INDY_SCHEMA_ID_EXAMPLE = IndySchemaId.EXAMPLE

ANONCREDS_SCHEMA_ID_VALIDATE = AnonCredsSchemaId()
ANONCREDS_SCHEMA_ID_EXAMPLE = AnonCredsSchemaId.EXAMPLE

INDY_CRED_DEF_ID_VALIDATE = IndyCredDefId()
INDY_CRED_DEF_ID_EXAMPLE = IndyCredDefId.EXAMPLE

ANONCREDS_CRED_DEF_ID_VALIDATE = AnonCredsCredDefId()
ANONCREDS_CRED_DEF_ID_EXAMPLE = AnonCredsCredDefId.EXAMPLE

INDY_REV_REG_ID_VALIDATE = IndyRevRegId()
INDY_REV_REG_ID_EXAMPLE = IndyRevRegId.EXAMPLE

ANONCREDS_REV_REG_ID_VALIDATE = AnonCredsRevRegId()
ANONCREDS_REV_REG_ID_EXAMPLE = AnonCredsRevRegId.EXAMPLE

INDY_CRED_REV_ID_VALIDATE = IndyCredRevId()
INDY_CRED_REV_ID_EXAMPLE = IndyCredRevId.EXAMPLE

ANONCREDS_CRED_REV_ID_VALIDATE = AnonCredsCredRevId()
ANONCREDS_CRED_REV_ID_EXAMPLE = AnonCredsCredRevId.EXAMPLE

MAJOR_MINOR_VERSION_VALIDATE = MajorMinorVersion()
MAJOR_MINOR_VERSION_EXAMPLE = MajorMinorVersion.EXAMPLE

PREDICATE_VALIDATE = Predicate()
PREDICATE_EXAMPLE = Predicate.EXAMPLE

ISO8601_DATETIME_VALIDATE = ISO8601DateTime()
ISO8601_DATETIME_EXAMPLE = ISO8601DateTime.EXAMPLE

RFC3339_DATETIME_VALIDATE = RFC3339DateTime()
RFC3339_DATETIME_EXAMPLE = RFC3339DateTime.EXAMPLE

INDY_WQL_VALIDATE = IndyWQL()
INDY_WQL_EXAMPLE = IndyWQL.EXAMPLE

INDY_EXTRA_WQL_VALIDATE = IndyExtraWQL()
INDY_EXTRA_WQL_EXAMPLE = IndyExtraWQL.EXAMPLE

BASE64_VALIDATE = Base64()
BASE64_EXAMPLE = Base64.EXAMPLE

BASE64URL_VALIDATE = Base64URL()
BASE64URL_EXAMPLE = Base64URL.EXAMPLE

BASE64URL_NO_PAD_VALIDATE = Base64URLNoPad()
BASE64URL_NO_PAD_EXAMPLE = Base64URLNoPad.EXAMPLE

SHA256_VALIDATE = SHA256Hash()
SHA256_EXAMPLE = SHA256Hash.EXAMPLE

BASE58_SHA256_HASH_VALIDATE = Base58SHA256Hash()
BASE58_SHA256_HASH_EXAMPLE = Base58SHA256Hash.EXAMPLE

UUID4_VALIDATE = UUIDFour()
UUID4_EXAMPLE = UUIDFour.EXAMPLE

ENDPOINT_VALIDATE = Endpoint()
ENDPOINT_EXAMPLE = Endpoint.EXAMPLE

ENDPOINT_TYPE_VALIDATE = EndpointType()
ENDPOINT_TYPE_EXAMPLE = EndpointType.EXAMPLE

CREDENTIAL_TYPE_VALIDATE = CredentialType()
CREDENTIAL_TYPE_EXAMPLE = CredentialType.EXAMPLE

CREDENTIAL_CONTEXT_VALIDATE = CredentialContext()
CREDENTIAL_CONTEXT_EXAMPLE = CredentialContext.EXAMPLE

URI_VALIDATE = Uri()
URI_EXAMPLE = Uri.EXAMPLE

CREDENTIAL_SUBJECT_VALIDATE = CredentialSubject()
CREDENTIAL_SUBJECT_EXAMPLE = CredentialSubject.EXAMPLE

CREDENTIAL_STATUS_VALIDATE = CredentialStatus()
CREDENTIAL_STATUS_EXAMPLE = CredentialStatus.EXAMPLE

PRESENTATION_TYPE_VALIDATE = PresentationType()
PRESENTATION_TYPE_EXAMPLE = PresentationType.EXAMPLE

INDY_OR_KEY_DID_VALIDATE = IndyOrKeyDID()
INDY_OR_KEY_DID_EXAMPLE = IndyOrKeyDID.EXAMPLE

ANONCREDS_DID_VALIDATE = AnonCredsDID()
ANONCREDS_DID_EXAMPLE = AnonCredsDID.EXAMPLE



"""Aries#0037 v1.0 presentation exchange information with non-secrets storage."""

import logging
from typing import Any, Mapping, Optional, Union

from marshmallow import fields, validate

from .....core.profile import ProfileSession
from .....indy.models.proof import IndyProof, IndyProofSchema
from .....indy.models.proof_request import IndyProofRequest, IndyProofRequestSchema
from .....messaging.models.base_record import BaseExchangeRecord, BaseExchangeSchema
from .....messaging.valid import UUID4_EXAMPLE
from .....storage.base import StorageError
from ..messages.presentation_proposal import (
    PresentationProposal,
    PresentationProposalSchema,
)
from ..messages.presentation_request import PresentationRequest, PresentationRequestSchema
from ..messages.presentation_webhook import V10PresentationExchangeWebhook
from . import UNENCRYPTED_TAGS

LOGGER = logging.getLogger(__name__)


class V10PresentationExchange(BaseExchangeRecord):
    """Represents an Aries#0037 v1.0 presentation exchange."""

    class Meta:
        """V10PresentationExchange metadata."""

        schema_class = "V10PresentationExchangeSchema"

    RECORD_TYPE = "presentation_exchange_v10"
    RECORD_ID_NAME = "presentation_exchange_id"
    RECORD_TOPIC = "present_proof"
    TAG_NAMES = {"~thread_id"} if UNENCRYPTED_TAGS else {"thread_id"}

    INITIATOR_SELF = "self"
    INITIATOR_EXTERNAL = "external"

    ROLE_PROVER = "prover"
    ROLE_VERIFIER = "verifier"

    STATE_PROPOSAL_SENT = "proposal_sent"
    STATE_PROPOSAL_RECEIVED = "proposal_received"
    STATE_REQUEST_SENT = "request_sent"
    STATE_REQUEST_RECEIVED = "request_received"
    STATE_PRESENTATION_SENT = "presentation_sent"
    STATE_PRESENTATION_RECEIVED = "presentation_received"
    STATE_VERIFIED = "verified"
    STATE_PRESENTATION_ACKED = "presentation_acked"
    STATE_ABANDONED = "abandoned"

    def __init__(
        self,
        *,
        presentation_exchange_id: Optional[str] = None,
        connection_id: Optional[str] = None,
        thread_id: Optional[str] = None,
        initiator: Optional[str] = None,
        role: Optional[str] = None,
        state: Optional[str] = None,
        presentation_proposal_dict: Union[
            PresentationProposal, Mapping
        ] = None,  # aries message: ..._dict for historic compat on all aries msgs
        presentation_request: Union[IndyProofRequest, Mapping] = None,  # indy proof req
        presentation_request_dict: Union[
            PresentationRequest, Mapping
        ] = None,  # aries message
        presentation: Union[IndyProof, Mapping] = None,  # indy proof
        verified: Optional[str] = None,
        verified_msgs: Optional[list] = None,
        auto_present: bool = False,
        auto_verify: bool = False,
        error_msg: Optional[str] = None,
        trace: bool = False,  # backward compat: BaseRecord.from_storage()
        auto_remove: bool = False,
        **kwargs,
    ):
        """Initialize a new PresentationExchange."""
        super().__init__(presentation_exchange_id, state, trace=trace, **kwargs)
        self.connection_id = connection_id
        self.thread_id = thread_id
        self.initiator = initiator
        self.role = role
        self.state = state
        self._presentation_proposal_dict = PresentationProposal.serde(
            presentation_proposal_dict
        )
        self._presentation_request = IndyProofRequest.serde(presentation_request)
        self._presentation_request_dict = PresentationRequest.serde(
            presentation_request_dict
        )
        self._presentation = IndyProof.serde(presentation)
        self.verified = verified
        self.verified_msgs = verified_msgs
        self.auto_present = auto_present
        self.auto_verify = auto_verify
        self.error_msg = error_msg
        self.auto_remove = auto_remove

    @property
    def presentation_exchange_id(self) -> str:
        """Accessor for the ID associated with this exchange."""
        return self._id

    @property
    def presentation_proposal_dict(self) -> PresentationProposal:
        """Accessor; get deserialized view."""
        return (
            None
            if self._presentation_proposal_dict is None
            else self._presentation_proposal_dict.de
        )

    @presentation_proposal_dict.setter
    def presentation_proposal_dict(self, value):
        """Setter; store de/serialized views."""
        self._presentation_proposal_dict = PresentationProposal.serde(value)

    @property
    def presentation_request(self) -> IndyProofRequest:
        """Accessor; get deserialized view."""
        return (
            None if self._presentation_request is None else self._presentation_request.de
        )

    @presentation_request.setter
    def presentation_request(self, value):
        """Setter; store de/serialized views."""
        self._presentation_request = IndyProofRequest.serde(value)

    @property
    def presentation_request_dict(self) -> PresentationRequest:
        """Accessor; get deserialized view."""
        return (
            None
            if self._presentation_request_dict is None
            else self._presentation_request_dict.de
        )

    @presentation_request_dict.setter
    def presentation_request_dict(self, value):
        """Setter; store de/serialized views."""
        self._presentation_request_dict = PresentationRequest.serde(value)

    @property
    def presentation(self) -> IndyProof:
        """Accessor; get deserialized view."""
        return None if self._presentation is None else self._presentation.de

    @presentation.setter
    def presentation(self, value):
        """Setter; store de/serialized views."""
        self._presentation = IndyProof.serde(value)

    async def save_error_state(
        self,
        session: ProfileSession,
        *,
        state: Optional[str] = None,
        reason: Optional[str] = None,
        log_params: Mapping[str, Any] = None,
        log_override: bool = False,
    ):
        """Save record error state if need be; log and swallow any storage error.

        Args:
            session: The profile session to use
            state: The state to set
            reason: A reason to add to the log
            log_params: Additional parameters to log
            log_override: Override configured logging regimen, print to stderr instead
        """

        if self._last_state == state:  # already done
            return

        self.state = state or V10PresentationExchange.STATE_ABANDONED
        if reason:
            self.error_msg = reason

        try:
            await self.save(
                session,
                reason=reason,
                log_params=log_params,
                log_override=log_override,
            )
        except StorageError as err:
            LOGGER.exception(err)

    # Override
    async def emit_event(self, session: ProfileSession, payload: Optional[Any] = None):
        """Emit an event.

        Args:
            session: The profile session to use
            payload: The event payload
        """

        if not self.RECORD_TOPIC:
            return

        if self.state:
            topic = f"{self.EVENT_NAMESPACE}::{self.RECORD_TOPIC}::{self.state}"
        else:
            topic = f"{self.EVENT_NAMESPACE}::{self.RECORD_TOPIC}"

        if session.profile.settings.get("debug.webhooks"):
            if not payload:
                payload = self.serialize()
        else:
            payload = V10PresentationExchangeWebhook(**self.__dict__)
            payload = payload.__dict__

        await session.profile.notify(topic, payload)

    @property
    def record_value(self) -> Mapping:
        """Accessor for the JSON record value generated for this credential exchange."""
        retval = {
            **{
                prop: getattr(self, prop)
                for prop in (
                    "connection_id",
                    "initiator",
                    "role",
                    "state",
                    "auto_present",
                    "auto_verify",
                    "error_msg",
                    "verified",
                    "verified_msgs",
                    "trace",
                    "auto_remove",
                )
            },
            **{
                prop: getattr(self, f"_{prop}").ser
                for prop in (
                    "presentation_proposal_dict",
                    "presentation_request",
                    "presentation_request_dict",
                    "presentation",
                )
                if getattr(self, prop) is not None
            },
        }
        return retval

    def __eq__(self, other: Any) -> bool:
        """Comparison between records."""
        return super().__eq__(other)


class V10PresentationExchangeSchema(BaseExchangeSchema):
    """Schema for de/serialization of v1.0 presentation exchange records."""

    class Meta:
        """V10PresentationExchangeSchema metadata."""

        model_class = V10PresentationExchange

    presentation_exchange_id = fields.Str(
        required=False,
        metadata={
            "description": "Presentation exchange identifier",
            "example": UUID4_EXAMPLE,
        },
    )
    connection_id = fields.Str(
        required=False,
        metadata={"description": "Connection identifier", "example": UUID4_EXAMPLE},
    )
    thread_id = fields.Str(
        required=False,
        metadata={"description": "Thread identifier", "example": UUID4_EXAMPLE},
    )
    initiator = fields.Str(
        required=False,
        validate=validate.OneOf(["self", "external"]),
        metadata={
            "description": "Present-proof exchange initiator: self or external",
            "example": V10PresentationExchange.INITIATOR_SELF,
        },
    )
    role = fields.Str(
        required=False,
        validate=validate.OneOf(["prover", "verifier"]),
        metadata={
            "description": "Present-proof exchange role: prover or verifier",
            "example": V10PresentationExchange.ROLE_PROVER,
        },
    )
    state = fields.Str(
        required=False,
        metadata={
            "description": "Present-proof exchange state",
            "example": V10PresentationExchange.STATE_VERIFIED,
        },
    )
    presentation_proposal_dict = fields.Nested(
        PresentationProposalSchema(),
        required=False,
        metadata={"description": "Presentation proposal message"},
    )
    presentation_request = fields.Nested(
        IndyProofRequestSchema(),
        required=False,
        metadata={
            "description": "(Indy) presentation request (also known as proof request)"
        },
    )
    presentation_request_dict = fields.Nested(
        PresentationRequestSchema(),
        required=False,
        metadata={"description": "Presentation request message"},
    )
    presentation = fields.Nested(
        IndyProofSchema(),
        required=False,
        metadata={"description": "(Indy) presentation (also known as proof)"},
    )
    verified = fields.Str(
        required=False,
        validate=validate.OneOf(["true", "false"]),
        metadata={
            "description": "Whether presentation is verified: true or false",
            "example": "true",
        },
    )
    verified_msgs = fields.List(
        fields.Str(
            required=False,
            metadata={"description": "Proof verification warning or error information"},
        ),
        required=False,
    )
    auto_present = fields.Bool(
        required=False,
        metadata={
            "description": "Prover choice to auto-present proof as verifier requests",
            "example": False,
        },
    )
    auto_verify = fields.Bool(
        required=False,
        metadata={"description": "Verifier choice to auto-verify proof presentation"},
    )
    error_msg = fields.Str(
        required=False,
        metadata={"description": "Error message", "example": "Invalid structure"},
    )
    auto_remove = fields.Bool(
        required=False,
        dump_default=True,
        metadata={
            "description": (
                "Verifier choice to remove this presentation exchange record when"
                " complete"
            ),
            "example": False,
        },
    )



from unittest import TestCase, mock

from ...core.error import BaseError
from .. import classloader as test_module
from ..classloader import ClassLoader, ClassNotFoundError, ModuleLoadError


class TestClassLoader(TestCase):
    def test_import_loaded(self):
        assert ClassLoader.load_module("unittest")

    def test_import_local(self):
        with mock.patch.object(test_module.sys, "modules", {}):
            assert (
                ClassLoader.load_module("acapy_agent.transport").__name__
                == "acapy_agent.transport"
            )

    def test_import_relative(self):
        with mock.patch.object(test_module.sys, "modules", {}):
            assert (
                ClassLoader.load_module("transport", "acapy_agent").__name__
                == "acapy_agent.transport"
            )
        with mock.patch.object(test_module.sys, "modules", {}):
            assert (
                ClassLoader.load_module(".transport", "acapy_agent").__name__
                == "acapy_agent.transport"
            )
        with mock.patch.object(test_module.sys, "modules", {}):
            assert (
                ClassLoader.load_module("..transport", "acapy_agent.config").__name__
                == "acapy_agent.transport"
            )

    def test_import_missing(self):
        with mock.patch.object(test_module.sys, "modules", {}):
            assert ClassLoader.load_module("acapy_agent.not") is None
        with mock.patch.object(test_module.sys, "modules", {}):
            assert ClassLoader.load_module("acapy_agent.not.a-module") is None
        with mock.patch.object(test_module.sys, "modules", {}):
            assert ClassLoader.load_module("acapy_agent", "not.a-module") is None

    def test_import_error(self):
        with (
            mock.patch.object(
                test_module, "import_module", autospec=True
            ) as import_module,
            mock.patch.object(test_module.sys, "modules", {}),
        ):
            import_module.side_effect = ModuleNotFoundError
            with self.assertRaises(ModuleLoadError):
                ClassLoader.load_module("acapy_agent.config")

    def test_load_class(self):
        assert ClassLoader.load_class("TestCase", "unittest") is TestCase
        assert ClassLoader.load_class("unittest.TestCase") is TestCase

    def test_load_class_missing(self):
        with self.assertRaises(ClassNotFoundError):
            # with no default module
            assert ClassLoader.load_class("NotAClass")
        with self.assertRaises(ClassNotFoundError):
            assert ClassLoader.load_class("acapy_agent.NotAClass")
        with self.assertRaises(ClassNotFoundError):
            assert ClassLoader.load_class("not-a-module.NotAClass")
        with self.assertRaises(ClassNotFoundError):
            # should be a string, not a type
            assert ClassLoader.load_class("acapy_agent.version.__version__")

    def test_load_subclass(self):
        assert ClassLoader.load_subclass_of(BaseError, "acapy_agent.config.base")

    def test_load_subclass_missing(self):
        with self.assertRaises(ClassNotFoundError):
            assert ClassLoader.load_subclass_of(TestCase, "acapy_agent.config.base")
        with self.assertRaises(ClassNotFoundError):
            assert ClassLoader.load_subclass_of(TestCase, "acapy_agent.not-a-module")

    def test_scan_packages(self):
        pkgs = ClassLoader.scan_subpackages("acapy_agent")
        assert "acapy_agent.transport" in pkgs
        pkgs = ClassLoader.scan_subpackages("acapy_agent.transport")
        assert "acapy_agent.transport.inbound" in pkgs

    def test_scan_packages_missing(self):
        with self.assertRaises(ModuleLoadError):
            ClassLoader.scan_subpackages("acapy_agent.not-a-module")



from ....vc.tests.contexts import (
    CITIZENSHIP_V1,
    CREDENTIALS_V1,
    EXAMPLES_V1,
    ODRL,
    SCHEMA_ORG,
    SECURITY_V1,
    SECURITY_V2,
)
from . import (
    TEST_EURO_HEALTH,
    TEST_SIGN_OBJ0,
    TEST_SIGN_OBJ1,
    TEST_VERIFY_ERROR,
    TEST_VERIFY_OBJ0,
    TEST_VERIFY_OBJ1,
)

DOCUMENTS = {
    TEST_SIGN_OBJ0["doc"]["id"]: TEST_SIGN_OBJ0["doc"],
    TEST_SIGN_OBJ1["doc"]["id"]: TEST_SIGN_OBJ1["doc"],
    TEST_VERIFY_ERROR["doc"]["id"]: TEST_VERIFY_ERROR["doc"],
    TEST_VERIFY_OBJ0["doc"]["id"]: TEST_VERIFY_OBJ0["doc"],
    TEST_VERIFY_OBJ1["doc"]["id"]: TEST_VERIFY_OBJ1["doc"],
    "https://w3id.org/citizenship/v1": CITIZENSHIP_V1,
    "https://www.w3.org/2018/credentials/v1": CREDENTIALS_V1,
    "https://www.w3.org/2018/credentials/examples/v1": EXAMPLES_V1,
    "https://www.w3.org/ns/odrl.jsonld": ODRL,
    "http://schema.org/": SCHEMA_ORG,
    "https://w3id.org/security/v1": SECURITY_V1,
    "https://w3id.org/security/v2": SECURITY_V2,
    (
        "https://essif-lab.pages.grnet.gr/interoperability/"
        "eidas-generic-use-case/contexts/ehic-v1.jsonld"
    ): TEST_EURO_HEALTH,
}


def custom_document_loader(url: str, options: dict):
    # Check if full url (with fragments is in document map)
    if url in DOCUMENTS:
        return {
            "contentType": "application/ld+json",
            "contextUrl": None,
            "document": DOCUMENTS[url],
            "documentUrl": url,
        }

    # Otherwise look if it is present without fragment
    without_fragment = url.split("#")[0]
    if without_fragment in DOCUMENTS:
        return {
            "contentType": "application/ld+json",
            "contextUrl": None,
            "document": DOCUMENTS[without_fragment],
            "documentUrl": url,
        }

    raise Exception(f"No custom context support for {url}")



"""A credential offer content message."""

from typing import Optional, Sequence

from marshmallow import EXCLUDE, fields

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.decorators.attach_decorator import (
    AttachDecorator,
    AttachDecoratorSchema,
)
from ..message_types import ATTACH_DECO_IDS, CREDENTIAL_OFFER, PROTOCOL_PACKAGE
from .inner.credential_preview import CredentialPreview, CredentialPreviewSchema

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers.credential_offer_handler.CredentialOfferHandler"
)


class CredentialOffer(AgentMessage):
    """Class representing a credential offer."""

    class Meta:
        """CredentialOffer metadata."""

        handler_class = HANDLER_CLASS
        schema_class = "CredentialOfferSchema"
        message_type = CREDENTIAL_OFFER

    def __init__(
        self,
        _id: Optional[str] = None,
        *,
        comment: Optional[str] = None,
        credential_preview: Optional[CredentialPreview] = None,
        offers_attach: Sequence[AttachDecorator] = None,
        **kwargs,
    ):
        """Initialize credential offer object.

        Args:
            comment: optional human-readable comment
            credential_preview: credential preview
            offers_attach: list of offer attachments
            kwargs: additional key-value arguments to map into message class properties

        """
        super().__init__(_id=_id, **kwargs)
        self.comment = comment
        self.credential_preview = credential_preview
        self.offers_attach = list(offers_attach) if offers_attach else []

    def indy_offer(self, index: int = 0) -> dict:
        """Retrieve and decode indy offer from attachment.

        Args:
            index: ordinal in attachment list to decode and return
                (typically, list has length 1)

        """
        return self.offers_attach[index].content

    @classmethod
    def wrap_indy_offer(cls, indy_offer: dict) -> AttachDecorator:
        """Convert an indy credential offer to an attachment decorator."""
        return AttachDecorator.data_base64(
            mapping=indy_offer, ident=ATTACH_DECO_IDS[CREDENTIAL_OFFER]
        )


class CredentialOfferSchema(AgentMessageSchema):
    """Credential offer schema."""

    class Meta:
        """Credential offer schema metadata."""

        model_class = CredentialOffer
        unknown = EXCLUDE

    comment = fields.Str(
        required=False,
        allow_none=True,
        metadata={"description": "Human-readable comment"},
    )
    credential_preview = fields.Nested(CredentialPreviewSchema, required=False)
    offers_attach = fields.Nested(
        AttachDecoratorSchema, required=True, many=True, data_key="offers~attach"
    )



"""Test handler for keylist-update-response message."""

from functools import partial
from typing import AsyncGenerator
from unittest import IsolatedAsyncioTestCase, mock

import pytest

from ......connections.models.conn_record import ConnRecord
from ......core.event_bus import EventBus, MockEventBus
from ......messaging.base_handler import HandlerException
from ......messaging.request_context import RequestContext
from ......messaging.responder import MockResponder
from ......utils.testing import create_test_profile
from ...manager import MediationManager
from ...messages.inner.keylist_update_rule import KeylistUpdateRule
from ...messages.inner.keylist_updated import KeylistUpdated
from ...messages.keylist_update_response import KeylistUpdateResponse
from ...route_manager import RouteManager
from ...tests.test_route_manager import MockRouteManager
from ..keylist_update_response_handler import KeylistUpdateResponseHandler

TEST_CONN_ID = "conn-id"
TEST_THREAD_ID = "thread-id"
TEST_VERKEY = "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
TEST_ROUTE_VERKEY = "did:key:z6MknxTj6Zj1VrDWc1ofaZtmCVv2zNXpD58Xup4ijDGoQhya"


class TestKeylistUpdateResponseHandler(IsolatedAsyncioTestCase):
    """Test handler for keylist-update-response message."""

    async def asyncSetUp(self):
        """Setup test dependencies."""
        self.context = RequestContext.test_context(await create_test_profile())
        self.updated = [
            KeylistUpdated(
                recipient_key=TEST_VERKEY,
                action=KeylistUpdateRule.RULE_ADD,
                result=KeylistUpdated.RESULT_SUCCESS,
            )
        ]
        self.context.message = KeylistUpdateResponse(updated=self.updated)
        self.context.connection_ready = True
        self.context.connection_record = ConnRecord(connection_id=TEST_CONN_ID)
        self.mock_event_bus = MockEventBus()
        self.context.profile.context.injector.bind_instance(EventBus, self.mock_event_bus)
        self.route_manager = MockRouteManager()
        self.context.profile.context.injector.bind_instance(
            RouteManager, self.route_manager
        )

    async def test_handler_no_active_connection(self):
        handler, responder = KeylistUpdateResponseHandler(), MockResponder()
        self.context.connection_ready = False
        with pytest.raises(HandlerException):
            await handler.handle(self.context, responder)

    async def test_handler(self):
        handler, responder = KeylistUpdateResponseHandler(), MockResponder()
        with (
            mock.patch.object(MediationManager, "store_update_results") as mock_store,
            mock.patch.object(handler, "notify_keylist_updated") as mock_notify,
        ):
            await handler.handle(self.context, responder)
            mock_store.assert_called_once_with(TEST_CONN_ID, self.updated)
            mock_notify.assert_called_once_with(
                self.context.profile, TEST_CONN_ID, self.context.message
            )

    async def test_notify_keylist_updated(self):
        """test notify_keylist_updated."""
        handler = KeylistUpdateResponseHandler()

        async def _result_generator():
            yield ConnRecord(connection_id="conn_id_1")
            yield ConnRecord(connection_id="conn_id_2")

        async def _retrieve_by_invitation_key(generator: AsyncGenerator, *args, **kwargs):
            return await generator.__anext__()

        with mock.patch.object(
            self.route_manager,
            "connection_from_recipient_key",
            partial(_retrieve_by_invitation_key, _result_generator()),
        ):
            response = KeylistUpdateResponse(
                updated=[
                    KeylistUpdated(
                        recipient_key=TEST_ROUTE_VERKEY,
                        action=KeylistUpdateRule.RULE_ADD,
                        result=KeylistUpdated.RESULT_SUCCESS,
                    ),
                    KeylistUpdated(
                        recipient_key=TEST_VERKEY,
                        action=KeylistUpdateRule.RULE_REMOVE,
                        result=KeylistUpdated.RESULT_SUCCESS,
                    ),
                ],
            )

            response.assign_thread_id(TEST_THREAD_ID)
            await handler.notify_keylist_updated(
                self.context.profile, TEST_CONN_ID, response
            )
        assert self.mock_event_bus.events
        assert (
            self.mock_event_bus.events[0][1].topic
            == MediationManager.KEYLIST_UPDATED_EVENT
        )
        assert self.mock_event_bus.events[0][1].payload == {
            "connection_id": TEST_CONN_ID,
            "thread_id": TEST_THREAD_ID,
            "updated": [result.serialize() for result in response.updated],
            "mediated_connections": {
                TEST_ROUTE_VERKEY: "conn_id_1",
                TEST_VERKEY: "conn_id_2",
            },
        }

    async def test_notify_keylist_updated_x_unknown_recip_key(self):
        """test notify_keylist_updated."""
        handler = KeylistUpdateResponseHandler()
        response = KeylistUpdateResponse(
            updated=[
                KeylistUpdated(
                    recipient_key=TEST_ROUTE_VERKEY,
                    action=KeylistUpdateRule.RULE_ADD,
                    result=KeylistUpdated.RESULT_SUCCESS,
                ),
                KeylistUpdated(
                    recipient_key=TEST_VERKEY,
                    action=KeylistUpdateRule.RULE_REMOVE,
                    result=KeylistUpdated.RESULT_SUCCESS,
                ),
            ],
        )

        response.assign_thread_id(TEST_THREAD_ID)
        with pytest.raises(HandlerException):
            await handler.notify_keylist_updated(
                self.context.profile, TEST_CONN_ID, response
            )



"""V2.0 issue-credential vc_di credential format handler.

indy compatible, attachment is a valid verifiable credential
"""

import datetime
import json
import logging
from typing import Mapping, Optional, Tuple

from anoncreds import W3cCredential
from marshmallow import RAISE

from ......anoncreds.holder import AnonCredsHolder, AnonCredsHolderError
from ......anoncreds.issuer import AnonCredsIssuer
from ......anoncreds.registry import AnonCredsRegistry
from ......anoncreds.revocation import AnonCredsRevocation
from ......cache.base import BaseCache
from ......ledger.base import BaseLedger
from ......ledger.multiple_ledger.ledger_requests_executor import (
    GET_CRED_DEF,
    IndyLedgerRequestsExecutor,
)
from ......messaging.credential_definitions.util import (
    CRED_DEF_SENT_RECORD_TYPE,
    CredDefQueryStringSchema,
)
from ......messaging.decorators.attach_decorator import AttachDecorator
from ......multitenant.base import BaseMultitenantManager
from ......revocation_anoncreds.models.issuer_cred_rev_record import IssuerCredRevRecord
from ......storage.base import BaseStorage
from ......vc.vc_ld import VerifiableCredential
from ......wallet.base import BaseWallet
from ...message_types import (
    ATTACHMENT_FORMAT,
    CRED_20_ISSUE,
    CRED_20_OFFER,
    CRED_20_PROPOSAL,
    CRED_20_REQUEST,
)
from ...messages.cred_format import V20CredFormat
from ...messages.cred_issue import V20CredIssue
from ...messages.cred_offer import V20CredOffer
from ...messages.cred_proposal import V20CredProposal
from ...messages.cred_request import V20CredRequest
from ...models.cred_ex_record import V20CredExRecord
from ...models.detail.indy import V20CredExRecordIndy
from ..handler import CredFormatAttachment, V20CredFormatError, V20CredFormatHandler
from .models.cred import VCDIIndyCredentialSchema
from .models.cred_offer import (
    AnonCredsLinkSecret,
    BindingMethod,
    DidcommSignedAttachment,
    VCDICredAbstract,
    VCDICredAbstractSchema,
)
from .models.cred_request import (
    AnonCredsLinkSecretRequest,
    BindingProof,
    DidcommSignedAttachmentRequest,
    VCDICredRequest,
    VCDICredRequestSchema,
)

LOGGER = logging.getLogger(__name__)


class VCDICredFormatHandler(V20CredFormatHandler):
    """VCDI credential format handler."""

    format = V20CredFormat.Format.VC_DI

    @classmethod
    def validate_fields(cls, message_type: str, attachment_data: Mapping):
        """Validate attachment data for a specific message type.

        Uses marshmallow schemas to validate if format specific attachment data
        is valid for the specified message type. Only does structural and type
        checks, does not validate if .e.g. the issuer value is valid.

        Args:
            message_type (str): The message type to validate the attachment data for.
                Should be one of the message types as defined in message_types.py
            attachment_data (Mapping): [description]
                The attachment data to valide
        Raises:
            Exception: When the data is not valid.
        """
        mapping = {
            CRED_20_PROPOSAL: CredDefQueryStringSchema,
            CRED_20_OFFER: VCDICredAbstractSchema,
            CRED_20_REQUEST: VCDICredRequestSchema,
            CRED_20_ISSUE: VCDIIndyCredentialSchema,
        }

        # Get schema class
        Schema = mapping[message_type]

        # Validate, throw if not valid
        Schema(unknown=RAISE).load(attachment_data)

    async def get_detail_record(self, cred_ex_id: str) -> V20CredExRecordIndy:
        """Retrieve credential exchange detail record by cred_ex_id."""

        async with self.profile.session() as session:
            records = await VCDICredFormatHandler.format.detail.query_by_cred_ex_id(
                session, cred_ex_id
            )

        if len(records) > 1:
            LOGGER.warning(
                "Cred ex id %s has %d %s detail records: should be 1",
                cred_ex_id,
                len(records),
                VCDICredFormatHandler.format.api,
            )
        return records[0] if records else None

    async def _check_uniqueness(self, cred_ex_id: str):
        """Raise exception on evidence that cred ex already has cred issued to it."""
        async with self.profile.session() as session:
            exist = await VCDICredFormatHandler.format.detail.query_by_cred_ex_id(
                session, cred_ex_id
            )
        if exist:
            raise V20CredFormatError(
                f"{VCDICredFormatHandler.format.api} detail record already "
                f"exists for cred ex id {cred_ex_id}"
            )

    def get_format_identifier(self, message_type: str) -> str:
        """Get attachment format identifier for format and message combination.

        Args:
            message_type (str): Message type for which to return the format identifier
        Returns:
            str: Issue credential attachment format identifier
        """
        return ATTACHMENT_FORMAT[message_type][VCDICredFormatHandler.format.api]

    def get_format_data(self, message_type: str, data: dict) -> CredFormatAttachment:
        """Get credential format and attachment objects for use in cred ex messages.

        Returns a tuple of both credential format and attachment decorator for use
        in credential exchange messages. It looks up the correct format identifier and
        encodes the data as a base64 attachment.

        Args:
            message_type (str): The message type for which to return the cred format.
                Should be one of the message types defined in the message types file
            data (dict): The data to include in the attach decorator
        Returns:
            CredFormatAttachment: Credential format and attachment data objects
        """
        return (
            V20CredFormat(
                attach_id=VCDICredFormatHandler.format.api,
                format_=self.get_format_identifier(message_type),
            ),
            AttachDecorator.data_base64(data, ident=VCDICredFormatHandler.format.api),
        )

    async def _match_sent_cred_def_id(self, tag_query: Mapping[str, str]) -> str:
        """Return most recent matching id of cred def that agent sent to ledger."""

        async with self.profile.session() as session:
            storage = session.inject(BaseStorage)
            found = await storage.find_all_records(
                type_filter=CRED_DEF_SENT_RECORD_TYPE, tag_query=tag_query
            )
        if not found:
            raise V20CredFormatError(
                f"Issuer has no operable cred def for proposal spec {tag_query}"
            )
        return max(found, key=lambda r: int(r.tags["epoch"])).tags["cred_def_id"]

    async def create_proposal(
        self, cred_ex_record: V20CredExRecord, proposal_data: Mapping[str, str]
    ) -> Tuple[V20CredFormat, AttachDecorator]:
        """Create vc_di credential proposal."""
        if proposal_data is None:
            proposal_data = {}

        return self.get_format_data(CRED_20_PROPOSAL, proposal_data)

    async def receive_proposal(
        self,
        cred_ex_record: V20CredExRecord,
        cred_proposal_message: V20CredProposal,
    ) -> None:
        """Receive vcdi credential proposal.

        No custom handling is required for this step.
        """

    async def create_offer(
        self, cred_proposal_message: V20CredProposal
    ) -> CredFormatAttachment:
        """Create vcdi credential offer."""

        issuer = AnonCredsIssuer(self.profile)
        # TODO use the ledger registry in the anoncreds module,
        # or move the functionality into the ledger class.
        ledger = self.profile.inject(BaseLedger)
        cache = self.profile.inject_or(BaseCache)

        async with self.profile.session() as session:
            wallet = session.inject(BaseWallet)
            public_did_info = await wallet.get_public_did()
            public_did = public_did_info.did

        cred_def_id = await issuer.match_created_credential_definitions(
            **cred_proposal_message.attachment(VCDICredFormatHandler.format)
        )

        async def _create():
            offer_str = await issuer.create_credential_offer(cred_def_id)
            return json.loads(offer_str)

        # TODO use the ledger registry in the anoncreds module,
        # or move the functionality into the ledger class.
        multitenant_mgr = self.profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(self.profile)
        else:
            ledger_exec_inst = self.profile.inject(IndyLedgerRequestsExecutor)
        ledger = (
            await ledger_exec_inst.get_ledger_for_identifier(
                cred_def_id,
                txn_record_type=GET_CRED_DEF,
            )
        )[1]
        # TODO use the ledger registry in the anoncreds module,
        # or move the functionality into the ledger class.
        async with ledger:
            schema_id = await ledger.credential_definition_id2schema_id(cred_def_id)
            schema = await ledger.get_schema(schema_id)
        schema_attrs = set(schema["attrNames"])
        preview_attrs = set(cred_proposal_message.credential_preview.attr_dict())
        if preview_attrs != schema_attrs:
            raise V20CredFormatError(
                f"Preview attributes {preview_attrs} "
                f"mismatch corresponding schema attributes {schema_attrs}"
            )

        cred_offer = None
        cache_key = f"credential_offer::{cred_def_id}"

        if cache:
            async with cache.acquire(cache_key) as entry:
                if entry.result:
                    cred_offer = entry.result
                else:
                    cred_offer = await _create()
                    await entry.set_result(cred_offer, 3600)
        if not cred_offer:
            cred_offer = await _create()

        credential = VerifiableCredential(
            issuer=public_did,
            credential_subject=cred_proposal_message.credential_preview.attr_dict(),
            issuance_date=datetime.datetime.now(datetime.timezone.utc).isoformat(),
        )

        anoncreds_link_secret_instance = AnonCredsLinkSecret(
            cred_def_id=cred_offer["cred_def_id"],
            key_correctness_proof=cred_offer["key_correctness_proof"],
            nonce=cred_offer["nonce"],
        )

        didcomm_signed_attachment_instance = DidcommSignedAttachment(
            algs_supported=["EdDSA"],
            did_methods_supported=["key"],
            nonce=cred_offer["nonce"],
        )

        binding_method_instance = BindingMethod(
            anoncreds_link_secret=anoncreds_link_secret_instance,
            didcomm_signed_attachment=didcomm_signed_attachment_instance,
        )

        vcdi_cred_abstract = VCDICredAbstract(
            data_model_versions_supported=["1.1"],
            binding_required=True,
            binding_method=binding_method_instance,
            credential=credential,
        )

        cred_offer = self.get_format_data(
            CRED_20_OFFER, json.loads(vcdi_cred_abstract.to_json())
        )
        return cred_offer

    async def receive_offer(
        self, cred_ex_record: V20CredExRecord, cred_offer_message: V20CredOffer
    ) -> None:
        """Receive vcdi credential offer."""

    async def create_request(
        self, cred_ex_record: V20CredExRecord, request_data: Optional[Mapping] = None
    ) -> CredFormatAttachment:
        """Create vcdi credential request."""
        if cred_ex_record.state != V20CredExRecord.STATE_OFFER_RECEIVED:
            raise V20CredFormatError(
                "vcdi issue credential format cannot start from credential request"
            )

        await self._check_uniqueness(cred_ex_record.cred_ex_id)

        holder_did = request_data.get("holder_did") if request_data else None
        cred_offer = cred_ex_record.cred_offer.attachment(VCDICredFormatHandler.format)

        if (
            "anoncreds_link_secret" in cred_offer["binding_method"]
            and "nonce" not in cred_offer["binding_method"]["anoncreds_link_secret"]
        ):
            raise V20CredFormatError(
                "Missing nonce in credential offer with anoncreds link secret "
                "binding method"
            )

        nonce = cred_offer["binding_method"]["anoncreds_link_secret"]["nonce"]
        cred_def_id = cred_offer["binding_method"]["anoncreds_link_secret"]["cred_def_id"]

        ledger = self.profile.inject(BaseLedger)
        # TODO use the ledger registry in the anoncreds module,
        # or move the functionality into the ledger class.
        multitenant_mgr = self.profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(self.profile)
        else:
            ledger_exec_inst = self.profile.inject(IndyLedgerRequestsExecutor)
        ledger = (
            await ledger_exec_inst.get_ledger_for_identifier(
                cred_def_id,
                txn_record_type=GET_CRED_DEF,
            )
        )[1]

        async with ledger:
            schema_id = await ledger.credential_definition_id2schema_id(cred_def_id)

        async def _create():
            anoncreds_registry = self.profile.inject(AnonCredsRegistry)

            cred_def_result = await anoncreds_registry.get_credential_definition(
                self.profile, cred_def_id
            )

            legacy_offer = await self._prepare_legacy_offer(cred_offer, schema_id)

            holder = AnonCredsHolder(self.profile)
            request_json, metadata_json = await holder.create_credential_request(
                legacy_offer, cred_def_result.credential_definition, holder_did
            )

            return {
                "request": json.loads(request_json),
                "metadata": json.loads(metadata_json),
            }

        cache_key = f"credential_request::{cred_def_id}::{holder_did}::{nonce}"
        cred_req_result = None
        cache = self.profile.inject_or(BaseCache)
        if cache:
            async with cache.acquire(cache_key) as entry:
                if entry.result:
                    cred_req_result = entry.result
                else:
                    cred_req_result = await _create()
                    await entry.set_result(cred_req_result, 3600)
        if not cred_req_result:
            cred_req_result = await _create()
        detail_record = V20CredExRecordIndy(
            cred_ex_id=cred_ex_record.cred_ex_id,
            cred_request_metadata=cred_req_result["metadata"],
        )

        if "prover_did" in cred_req_result["request"]:
            entropy = cred_req_result["request"]["prover_did"]
        else:
            entropy = cred_req_result["request"]["entropy"]
        anoncreds_link_secret_instance = AnonCredsLinkSecretRequest(
            entropy=entropy,
            cred_def_id=cred_req_result["request"]["cred_def_id"],
            blinded_ms=cred_req_result["request"]["blinded_ms"],
            blinded_ms_correctness_proof=cred_req_result["request"][
                "blinded_ms_correctness_proof"
            ],
            nonce=cred_req_result["request"]["nonce"],
        )

        didcomm_signed_attachment_instance = DidcommSignedAttachmentRequest(
            attachment_id="test"
        )

        binding_proof_instance = BindingProof(
            anoncreds_link_secret=anoncreds_link_secret_instance,
            didcomm_signed_attachment=didcomm_signed_attachment_instance,
        )

        vcdi_cred_request = VCDICredRequest(
            data_model_version="2.0", binding_proof=binding_proof_instance
        )

        async with self.profile.session() as session:
            await detail_record.save(session, reason="create v2.0 credential request")

        return self.get_format_data(
            CRED_20_REQUEST, json.loads(vcdi_cred_request.to_json())
        )

    async def receive_request(
        self, cred_ex_record: V20CredExRecord, cred_request_message: V20CredRequest
    ) -> None:
        """Receive vcdi credential request."""
        if not cred_ex_record.cred_offer:
            raise V20CredFormatError(
                "vcdi issue credential format cannot start from credential request"
            )

    async def issue_credential(
        self, cred_ex_record: V20CredExRecord, retries: int = 5
    ) -> CredFormatAttachment:
        """Issue vcdi credential."""
        await self._check_uniqueness(cred_ex_record.cred_ex_id)

        cred_offer = cred_ex_record.cred_offer.attachment(VCDICredFormatHandler.format)
        cred_request = cred_ex_record.cred_request.attachment(
            VCDICredFormatHandler.format
        )
        cred_values = cred_ex_record.cred_offer.credential_preview.attr_dict(decode=False)

        cred_def_id = cred_offer["binding_method"]["anoncreds_link_secret"]["cred_def_id"]

        ledger = self.profile.inject(BaseLedger)
        # TODO use the ledger registry in the anoncreds module,
        # or move the functionality into the ledger class.
        multitenant_mgr = self.profile.inject_or(BaseMultitenantManager)
        if multitenant_mgr:
            ledger_exec_inst = IndyLedgerRequestsExecutor(self.profile)
        else:
            ledger_exec_inst = self.profile.inject(IndyLedgerRequestsExecutor)
        ledger = (
            await ledger_exec_inst.get_ledger_for_identifier(
                cred_def_id,
                txn_record_type=GET_CRED_DEF,
            )
        )[1]

        async with ledger:
            schema_id = await ledger.credential_definition_id2schema_id(cred_def_id)
            cred_def = await ledger.get_credential_definition(cred_def_id)
        revocable = True if cred_def["value"].get("revocation") else False

        legacy_offer = await self._prepare_legacy_offer(cred_offer, schema_id)
        legacy_request = await self._prepare_legacy_request(cred_request, cred_def_id)

        cred_rev_id = None
        rev_reg_def_id = None
        credential = None
        if revocable:
            issuer = AnonCredsRevocation(self.profile)
            (
                credential,
                cred_rev_id,
                rev_reg_def_id,
            ) = await issuer.create_credential_w3c(
                legacy_offer, legacy_request, cred_values
            )
        else:
            issuer = AnonCredsIssuer(self.profile)
            credential = await issuer.create_credential_w3c(
                legacy_offer, legacy_request, cred_values
            )

        vcdi_credential = {
            "credential": json.loads(credential),
        }

        result = self.get_format_data(CRED_20_ISSUE, vcdi_credential)

        async with self._profile.transaction() as txn:
            detail_record = V20CredExRecordIndy(
                cred_ex_id=cred_ex_record.cred_ex_id,
                rev_reg_id=rev_reg_def_id,
                cred_rev_id=cred_rev_id,
            )
            await detail_record.save(txn, reason="v2.0 issue credential")

            if cred_rev_id:
                issuer_cr_rec = IssuerCredRevRecord(
                    state=IssuerCredRevRecord.STATE_ISSUED,
                    cred_ex_id=cred_ex_record.cred_ex_id,
                    cred_ex_version=IssuerCredRevRecord.VERSION_2,
                    rev_reg_id=rev_reg_def_id,
                    cred_rev_id=cred_rev_id,
                )
                await issuer_cr_rec.save(
                    txn,
                    reason=(
                        "Created issuer cred rev record for "
                        f"rev reg id {rev_reg_def_id}, index {cred_rev_id}"
                    ),
                )
            await txn.commit()

        return result

    async def _prepare_legacy_offer(self, cred_offer: dict, schema_id: str) -> dict:
        """Convert current offer to legacy offer format."""
        return {
            "schema_id": schema_id,
            "cred_def_id": cred_offer["binding_method"]["anoncreds_link_secret"][
                "cred_def_id"
            ],
            "key_correctness_proof": cred_offer["binding_method"][
                "anoncreds_link_secret"
            ]["key_correctness_proof"],
            "nonce": cred_offer["binding_method"]["anoncreds_link_secret"]["nonce"],
        }

    async def _prepare_legacy_request(self, cred_request: dict, cred_def_id: str):
        return {
            "prover_did": cred_request["binding_proof"]["anoncreds_link_secret"][
                "entropy"
            ],
            "cred_def_id": cred_def_id,
            "blinded_ms": cred_request["binding_proof"]["anoncreds_link_secret"][
                "blinded_ms"
            ],
            "blinded_ms_correctness_proof": cred_request["binding_proof"][
                "anoncreds_link_secret"
            ]["blinded_ms_correctness_proof"],
            "nonce": cred_request["binding_proof"]["anoncreds_link_secret"]["nonce"],
        }

    async def receive_credential(
        self, cred_ex_record: V20CredExRecord, cred_issue_message: V20CredIssue
    ) -> None:
        """Receive vcdi credential.

        Validation is done in the store credential step.
        """

    async def store_credential(
        self, cred_ex_record: V20CredExRecord, cred_id: Optional[str] = None
    ) -> None:
        """Store vcdi credential."""
        cred = cred_ex_record.cred_issue.attachment(VCDICredFormatHandler.format)
        cred = cred["credential"]
        try:
            w3cred = W3cCredential.load(cred)
        except AnonCredsHolderError as e:
            LOGGER.error(f"Error receiving credential: {e.error_code} - {e.message}")
        rev_reg_def = None
        anoncreds_registry = self.profile.inject(AnonCredsRegistry)
        cred_def_result = await anoncreds_registry.get_credential_definition(
            self.profile, cred["proof"][0]["verificationMethod"]
        )
        rev_reg_id = None
        rev_reg_index = None
        if (
            w3cred.rev_reg_id != "None"
        ):  # String None because rev_reg_id property wrapped str()
            rev_reg_id = w3cred.rev_reg_id

            rev_reg_def_result = (
                await anoncreds_registry.get_revocation_registry_definition(
                    self.profile, rev_reg_id
                )
            )
            rev_reg_def = rev_reg_def_result.revocation_registry

        holder = AnonCredsHolder(self.profile)
        cred_offer_message = cred_ex_record.cred_offer
        mime_types = None
        if cred_offer_message and cred_offer_message.credential_preview:
            mime_types = cred_offer_message.credential_preview.mime_types() or None

        if rev_reg_def:
            revocation = AnonCredsRevocation(self.profile)
            await revocation.get_or_fetch_local_tails_path(rev_reg_def)
        try:
            detail_record = await self.get_detail_record(cred_ex_record.cred_ex_id)
            if detail_record is None:
                raise V20CredFormatError(
                    f"No credential exchange {VCDICredFormatHandler.format.aries} "
                    f"detail record found for cred ex id {cred_ex_record.cred_ex_id}"
                )
            cred_id_stored = await holder.store_credential_w3c(
                cred_def_result.credential_definition.serialize(),
                cred,
                detail_record.cred_request_metadata,
                mime_types,
                credential_id=cred_id,
                rev_reg_def=rev_reg_def.serialize() if rev_reg_def else None,
            )

            detail_record.cred_id_stored = cred_id_stored
            detail_record.rev_reg_id = rev_reg_id
            detail_record.cred_rev_id = rev_reg_index

            async with self.profile.session() as session:
                # Store detail record, emit event
                await detail_record.save(
                    session, reason="store credential v2.0", event=True
                )
        except AnonCredsHolderError as e:
            LOGGER.error(f"Error storing credential: {e.error_code} - {e.message}")
            raise e



TEST_VC_DOCUMENT_NESTED_PROOF_BBS = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
        "https://w3id.org/security/bbs/v1",
    ],
    "id": "https://example.gov/credentials/3732",
    "type": ["UniversityDegreeCredential", "VerifiableCredential"],
    "credentialSubject": {
        "id": "did:example:489398593",
        "college": "Contoso University",
        "degree": {
            "id": "urn:bnid:_:c14n0",
            "type": "BachelorDegree",
            "name": "Bachelor of Science and Arts",
            "degreeType": "Underwater Basket Weaving",
        },
    },
    "issuanceDate": "2020-03-10T04:24:12.164Z",
    "issuer": "did:example:489398593",
    "proof": {
        "type": "BbsBlsSignatureProof2020",
        "created": "2021-04-26T17:55:05Z",
        "nonce": "Vy7Mup3vFgTcs0NxggmHYMfIf7+ykk/M1a0+wRNjcGukpe/C2Rl7WvN1DVxCg8gvhng=",
        "proofPurpose": "assertionMethod",
        "proofValue": "AA4//7SYcnewiCvH+CG/CwNY4rXdpfZ/9j79GASJFWa6C9jwBzQenfeXtuMBvHCDvmpTqI5Dtf/7O938CV0qEtUTquPrT6stNxnCfJSPYmM78d48LJidPoLhdRjT+QNfj5sM647odR5b9JOMmW43yMUAxkb+A9bGNdixG5OqK3xLSuRHDkTIi1Lu6/x4xgOeR5KALAAAAHSYATbHEjd2sdxY9n7lzdWkta25moSE1UNZBO+B+nSysYlQYO9H9jX0zRrzUaKDnSAAAAACX8oBf2PCWC3U6jtaoRqO2pAmc4nXO2tqsnVm6zNBUmEQlWf32/Cl3VvyB2q9XBGagPDLiyxuDcSkoQRJX4D/M4IRx8LVMhIYPT4O4mDejKuFqFQN1PiVN4W+eR0I8KZ0Jq8QP97CM8/6YFg+fyuxdQAAAAJJotlRjal+sxbNpyVHU70FDv1icZyEFJwGm7HnFFh/iAqjP69KB37eou5Xt1TkM6fgK65gu0e4jTTALpk5tie3",
        "verificationMethod": "did:example:489398593#test",
    },
}



"""Represents a coordinate-mediation problem report message."""

import logging
from enum import Enum

from marshmallow import EXCLUDE, ValidationError, validates_schema

from ....problem_report.v1_0.message import ProblemReport, ProblemReportSchema
from ..message_types import PROBLEM_REPORT, PROTOCOL_PACKAGE

HANDLER_CLASS = (
    f"{PROTOCOL_PACKAGE}.handlers.problem_report_handler.CMProblemReportHandler"
)

LOGGER = logging.getLogger(__name__)


class ProblemReportReason(Enum):
    """Supported reason codes."""

    MEDIATION_NOT_GRANTED = "mediation_not_granted"
    MEDIATION_REQUEST_REPEAT = "request_already_exists_from_connection"


class CMProblemReport(ProblemReport):
    """Base class representing a coordinate mediation problem report message."""

    class Meta:
        """CMProblemReport metadata."""

        handler_class = HANDLER_CLASS
        message_type = PROBLEM_REPORT
        schema_class = "CMProblemReportSchema"

    def __init__(self, *args, **kwargs):
        """Initialize a ProblemReport message instance."""
        super().__init__(*args, **kwargs)


class CMProblemReportSchema(ProblemReportSchema):
    """Schema for ProblemReport base class."""

    class Meta:
        """Metadata for problem report schema."""

        model_class = CMProblemReport
        unknown = EXCLUDE

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields."""

        if not data.get("description", {}).get("code", ""):
            raise ValidationError("Value for description.code must be present")
        elif data.get("description", {}).get("code", "") not in [
            prr.value for prr in ProblemReportReason
        ]:
            locales = list(data.get("description").keys())
            locales.remove("code")
            LOGGER.warning(
                "Unexpected error code received.\n"
                f"Code: {data.get('description').get('code')}, "
                f"Description: {data.get('description').get(locales[0])}"
            )



"""An invitation content message."""

from enum import Enum
from typing import NamedTuple, Optional, Sequence, Set, Text, Union
from urllib.parse import parse_qs, urljoin, urlparse

from marshmallow import EXCLUDE, ValidationError, fields, post_dump, validates_schema

from .....messaging.agent_message import AgentMessage, AgentMessageSchema
from .....messaging.decorators.attach_decorator import (
    AttachDecorator,
    AttachDecoratorSchema,
)
from .....messaging.valid import DIDValidation
from .....wallet.util import b64_to_bytes, bytes_to_b64
from ....didcomm_prefix import DIDCommPrefix
from ....didexchange.v1_0.message_types import ARIES_PROTOCOL as DIDEX_1_1
from ....didexchange.v1_0.message_types import DIDEX_1_0
from ..message_types import DEFAULT_VERSION, INVITATION
from .service import Service


class HSProtoSpec(NamedTuple):
    """Handshake protocol specification."""

    name: str
    aka: Set[str]


class HSProto(Enum):
    """Handshake protocol enum for invitation message."""

    RFC160 = HSProtoSpec(
        "connections/1.0",
        {"connection", "connections", "conn", "conns", "rfc160", "160", "old"},
    )
    RFC23 = HSProtoSpec(
        DIDEX_1_0,
        {
            "https://didcomm.org/didexchange/1.0",
            "didexchange/1.0",
            "didexchange",
            "did-exchange",
            "didx",
            "didex",
            "rfc23",
            "rfc-23",
            "23",
            "new",
        },
    )
    DIDEX_1_1 = HSProtoSpec(
        DIDEX_1_1,
        {
            "https://didcomm.org/didexchange/1.1",
            "didexchange/1.1",
        },
    )

    @classmethod
    def get(cls, label: Union[str, "HSProto"]) -> Optional["HSProto"]:
        """Get handshake protocol enum for label."""

        if isinstance(label, str):
            for hsp in HSProto:
                if DIDCommPrefix.unqualify(label) == hsp.name or label.lower() in hsp.aka:
                    return hsp

        elif isinstance(label, HSProto):
            return label

        return None

    @property
    def name(self) -> str:
        """Accessor for name."""
        return self.value.name

    @property
    def aka(self) -> Set[str]:
        """Accessor for also-known-as."""
        return self.value.aka


class ServiceOrDIDField(fields.Field):
    """DIDComm Service object or DID string field for Marshmallow."""

    def _serialize(self, value, attr, obj, **kwargs):
        if isinstance(value, Service):
            return value.serialize()
        return value

    def _deserialize(self, value, attr, data, **kwargs):
        if isinstance(value, dict):
            return Service.deserialize(value)
        elif isinstance(value, Service):
            return value
        elif isinstance(value, str):
            if not DIDValidation.PATTERN.match(value):
                raise ValidationError(
                    "Service item must be a valid decentralized identifier (DID)"
                )
            return value
        raise ValidationError(
            "Service item must be a valid decentralized identifier (DID) or object"
        )


class InvitationMessage(AgentMessage):
    """Class representing an out of band invitation message."""

    class Meta:
        """InvitationMessage metadata."""

        schema_class = "InvitationMessageSchema"
        message_type = INVITATION

    def __init__(
        self,
        *,
        label: Optional[str] = None,
        image_url: Optional[str] = None,
        handshake_protocols: Optional[Sequence[Text]] = None,
        requests_attach: Optional[Sequence[AttachDecorator]] = None,
        services: Optional[Sequence[Union[Service, Text]]] = None,
        accept: Optional[Sequence[Text]] = None,
        version: str = DEFAULT_VERSION,
        msg_type: Optional[Text] = None,
        goal_code: Optional[Text] = None,
        goal: Optional[Text] = None,
        **kwargs,
    ):
        """Initialize invitation message object.

        Args:
            label (Optional[str]): The label for the invitation.
            image_url (Optional[str]): The URL of an image associated with the invitation.
            handshake_protocols (Optional[Sequence[Text]]): The supported handshake
                protocols.
            requests_attach (Optional[Sequence[AttachDecorator]]): The request
                attachments.
            services (Optional[Sequence[Union[Service, Text]]]): The services associated
                with the invitation.
            accept (Optional[Sequence[Text]]): The accepted protocols.
            version (str): The version of the invitation message.
            msg_type (Optional[Text]): The type of the invitation message.
            goal_code (Optional[Text]): The goal code.
            goal (Optional[Text]): The goal.
            kwargs: Additional keyword arguments.

        """
        super().__init__(_type=msg_type, _version=version, **kwargs)
        self.label = label
        self.image_url = image_url
        self.handshake_protocols = (
            list(handshake_protocols) if handshake_protocols else []
        )
        self.requests_attach = list(requests_attach) if requests_attach else []
        self.services = services
        self.accept = accept
        self.goal_code = goal_code
        self.goal = goal

    @classmethod
    def wrap_message(cls, message: dict) -> AttachDecorator:
        """Convert an aries message to an attachment decorator."""
        return AttachDecorator.data_json(mapping=message, ident="request-0")

    def to_url(self, base_url: Optional[str] = None) -> str:
        """Convert an invitation message to URL format for sharing.

        Returns:
            An invite url

        """
        c_json = self.to_json()
        oob = bytes_to_b64(c_json.encode("ascii"), urlsafe=True, pad=False)
        endpoint = None
        if not base_url:
            for service_item in self.services:
                if isinstance(service_item, Service):
                    endpoint = service_item.service_endpoint
                    break
        result = urljoin(
            (base_url if base_url else endpoint),
            "?oob={}".format(oob),
        )
        return result

    @classmethod
    def from_url(cls, url: str) -> "InvitationMessage":
        """Parse a URL-encoded invitation into an `InvitationMessage` instance.

        Args:
            url: Url to decode

        Returns:
            An `InvitationMessage` object.

        """
        parts = urlparse(url)
        query = parse_qs(parts.query)
        if "oob" in query:
            oob = b64_to_bytes(query["oob"][0], urlsafe=True)
            return cls.from_json(oob)
        return None


class InvitationMessageSchema(AgentMessageSchema):
    """InvitationMessage schema."""

    class Meta:
        """InvitationMessage schema metadata."""

        model_class = InvitationMessage
        unknown = EXCLUDE

    _type = fields.Str(
        data_key="@type",
        required=False,
        metadata={
            "description": "Message type",
            "example": "https://didcomm.org/my-family/1.0/my-message-type",
        },
    )
    label = fields.Str(
        required=False, metadata={"description": "Optional label", "example": "Bob"}
    )
    image_url = fields.URL(
        data_key="imageUrl",
        required=False,
        allow_none=True,
        metadata={
            "description": "Optional image URL for out-of-band invitation",
            "example": "http://192.168.56.101/img/logo.jpg",
        },
    )
    handshake_protocols = fields.List(
        fields.Str(
            metadata={
                "description": "Handshake protocol",
                "example": DIDCommPrefix.qualify_current(HSProto.RFC23.name),
            }
        ),
        required=False,
    )
    accept = fields.List(
        fields.Str(),
        required=False,
        metadata={
            "example": ["didcomm/aip1", "didcomm/aip2;env=rfc19"],
            "description": "List of mime type in order of preference",
        },
    )
    requests_attach = fields.Nested(
        AttachDecoratorSchema,
        required=False,
        many=True,
        data_key="requests~attach",
        metadata={"description": "Optional request attachment"},
    )
    services = fields.List(
        ServiceOrDIDField(
            required=True,
            metadata={
                "description": (
                    "Either a DIDComm service object (as per RFC0067) or a DID string."
                )
            },
        ),
        metadata={
            "example": [
                {
                    "did": "WgWxqztrNooG92RXvxSTWv",
                    "id": "string",
                    "recipientKeys": [
                        "did:key:z6MkpTHR8VNsBxYAAWHut2Geadd9jSwuBV8xRoAnwWsdvktH"
                    ],
                    "routingKeys": [
                        "did:key:z6MkpTHR8VNsBxYAAWHut2Geadd9jSwuBV8xRoAnwWsdvktH"
                    ],
                    "serviceEndpoint": "http://192.168.56.101:8020",
                    "type": "string",
                },
                "did:sov:WgWxqztrNooG92RXvxSTWv",
            ]
        },
    )
    goal_code = fields.Str(
        required=False,
        metadata={
            "description": (
                "A self-attested code the receiver may want to display to the user or"
                " use in automatically deciding what to do with the out-of-band message"
            ),
            "example": "issue-vc",
        },
    )
    goal = fields.Str(
        required=False,
        metadata={
            "description": (
                "A self-attested string that the receiver may want to display to the"
                " user about the context-specific goal of the out-of-band message"
            ),
            "example": "To issue a Faber College Graduate credential",
        },
    )

    @validates_schema
    def validate_fields(self, data, **kwargs):
        """Validate schema fields.

        Args:
            data: The data to validate
            kwargs: Additional keyword arguments
        Raises:
            ValidationError: If any of the fields do not validate
        """
        handshake_protocols = data.get("handshake_protocols")
        requests_attach = data.get("requests_attach")
        if not handshake_protocols and not requests_attach:
            raise ValidationError(
                "Model must include non-empty "
                "handshake_protocols or requests~attach or both"
            )

        # services = data.get("services")
        # if not ((services and len(services) > 0)):
        #     raise ValidationError(
        #         "Model must include non-empty services array"
        #     )
        goal = data.get("goal")
        goal_code = data.get("goal_code")
        if goal and not goal_code:
            raise ValidationError("Model cannot have goal without goal_code")
        if goal_code and not goal:
            raise ValidationError("Model cannot have goal_code without goal")

    @post_dump
    def post_dump(self, data, **kwargs):
        """Post dump hook."""
        if "requests~attach" in data and not data["requests~attach"]:
            del data["requests~attach"]

        return data



from unittest import IsolatedAsyncioTestCase, TestCase, mock

from acapy_agent.protocols.coordinate_mediation.mediation_invite_store import (
    MediationInviteRecord,
    MediationInviteStore,
    NoDefaultMediationInviteException,
)
from acapy_agent.storage.base import BaseStorage
from acapy_agent.storage.error import StorageNotFoundError
from acapy_agent.storage.record import StorageRecord


def _storage_record_for(value: str, used: bool = False) -> StorageRecord:
    return StorageRecord(
        type=MediationInviteStore.INVITE_RECORD_CATEGORY,
        value=f"""{{"invite": "{value}", "used": {str(used).lower()}}}""",
        tags={},
        id=MediationInviteStore.MEDIATION_INVITE_ID,
    )


class TestMediationInviteRecord(TestCase):
    def test_to_json_should_dump_record(self):
        # given
        invite_record = MediationInviteRecord("some_invite", True)

        # when
        json_record = invite_record.to_json()

        # then
        assert json_record == """{"invite": "some_invite", "used": true}"""

    def test_from_json_should_create_record_from_json(self):
        # given
        json_record = """{"invite": "some_invite", "used": true}"""

        # when
        record = MediationInviteRecord.from_json(json_record)

        # then
        assert record == MediationInviteRecord("some_invite", True)

    def test_unused_should_create_unused_record(self):
        # when - then
        assert not MediationInviteRecord.unused("some_other_invite").used


class TestMediationInviteStore(IsolatedAsyncioTestCase):
    def setUp(self):
        self.storage = mock.MagicMock(spec=BaseStorage)
        self.mediation_invite_store = MediationInviteStore(self.storage)

    async def test_store_create_record_to_store_mediation_invite_when_no_record_exists(
        self,
    ):
        # given
        mediation_invite_url = "somepla.ce:4242/alongandunreadablebase64payload"
        self.storage.get_record.side_effect = StorageNotFoundError

        expected_updated_record = MediationInviteRecord.unused(mediation_invite_url)

        # when
        stored_invite = await self.mediation_invite_store.store(
            MediationInviteRecord.unused(mediation_invite_url)
        )

        # then
        self.storage.add_record.assert_called_with(
            _storage_record_for(mediation_invite_url)
        )
        assert stored_invite == expected_updated_record

    async def test_store_should_update_record_when_a_mediation_invite_record_exists(
        self,
    ):
        # given
        stored_record = _storage_record_for("some old url")
        mediation_invite_url = "somepla.ce:4242/alongandunreadablebase64payload"
        self.storage.get_record.return_value = stored_record

        expected_updated_record = MediationInviteRecord.unused(mediation_invite_url)

        # when
        stored_invite = await self.mediation_invite_store.store(
            MediationInviteRecord.unused(mediation_invite_url)
        )

        # then
        self.storage.update_record.assert_called_with(
            stored_record, expected_updated_record.to_json(), tags=stored_record.tags
        )

        assert stored_invite == expected_updated_record

    async def test_mark_default_invite_as_used_should_mark_stored_invite(self):
        # given
        stored_record = _storage_record_for("some old url")
        self.storage.get_record.return_value = stored_record

        # when
        updated_invite_record = (
            await self.mediation_invite_store.mark_default_invite_as_used()
        )

        # then
        assert updated_invite_record.used

    async def test_mark_default_invite_as_used_should_raise_when_no_invite(self):
        # given
        self.storage.get_record.return_value = None

        # when - then
        with self.assertRaises(NoDefaultMediationInviteException):
            await self.mediation_invite_store.mark_default_invite_as_used()

    async def test_get_mediation_invite_record_returns_none_when_no_invite_available(
        self,
    ):
        # given
        self.storage.get_record.side_effect = StorageNotFoundError

        # when
        invite = await self.mediation_invite_store.get_mediation_invite_record(None)

        # then
        assert invite is None

    async def test_get_mediation_invite_returns_stored_record_when_no_invite_provided(
        self,
    ):
        # given
        stored_record = _storage_record_for(
            "somepla.ce:4242/alongandunreadablebase64payload"
        )
        expected_invite = MediationInviteRecord.from_json(stored_record.value)
        self.storage.get_record.return_value = stored_record

        # when
        invite = await self.mediation_invite_store.get_mediation_invite_record(None)

        # then
        assert invite == expected_invite

    async def test_get_mediation_invite_stores_and_returns_provided_invite_if_none_stored(
        self,
    ):
        # given
        expected_invite = MediationInviteRecord.unused(
            "somepla.ce:4242/alongandunreadablebase64payload"
        )
        self.storage.get_record.return_value = None

        # when
        invite = await self.mediation_invite_store.get_mediation_invite_record(
            expected_invite.invite
        )

        # then
        assert invite == expected_invite
